"""Public interface managing the workflow for peer assessments.

The Peer Assessment Workflow API exposes all public actions required to complete
the workflow for a given submission.

"""


import logging

from django.db import DatabaseError, IntegrityError, transaction
from django.utils import timezone

from submissions import api as sub_api
from openassessment.assessment.errors import (PeerAssessmentInternalError, PeerAssessmentRequestError,
                                              PeerAssessmentWorkflowError)
from openassessment.assessment.models import (Assessment, AssessmentFeedback, AssessmentPart, InvalidRubricSelection,
                                              PeerWorkflow, PeerWorkflowItem)
from openassessment.assessment.serializers import (AssessmentFeedbackSerializer, InvalidRubric, RubricSerializer,
                                                   full_assessment_dict, rubric_from_dict, serialize_assessments)

logger = logging.getLogger("openassessment.assessment.api.peer")  # pylint: disable=invalid-name

PEER_TYPE = "PE"

FLEXIBLE_PEER_GRADING_REQUIRED_SUBMISSION_AGE_IN_DAYS = 7
FLEXIBLE_PEER_GRADING_GRADED_BY_PERCENTAGE = 30


def required_peer_grades(submission_uuid, peer_requirements):
    """
    Given a submission id, finds how many peer assessment required.

    Args:
        submission_uuid (str): The UUID of the submission being tracked.
        peer_requirements (dict): Dictionary with the key "must_grade" indicating
            the required number of submissions the student must grade
            and "enable_flexible_grading" indicating if flexible grading enabled.

    Returns:
        int
    """

    submission = sub_api.get_submission(submission_uuid)

    must_grade = peer_requirements["must_be_graded_by"]

    if peer_requirements.get("enable_flexible_grading"):

        # find how many days elapsed since subimitted
        days_elapsed = (timezone.now().date() - submission['submitted_at'].date()).days

        # check if flexible grading applies. if it does, then update must_grade
        if days_elapsed >= FLEXIBLE_PEER_GRADING_REQUIRED_SUBMISSION_AGE_IN_DAYS:
            must_grade = int(must_grade * FLEXIBLE_PEER_GRADING_GRADED_BY_PERCENTAGE / 100)
            if must_grade == 0:
                must_grade = 1

    return must_grade


def can_be_skipped(submission_uuid, peer_requirements):  # pylint: disable=unused-argument
    """
    Peer workflow step can be always skipped.

    Args:
        submission_uuid (str): The UUID of the submission being tracked.
        peer_requirements (dict): Dictionary with the key "must_grade" indicating
            the required number of submissions the student must grade.

    Returns:
        bool
    """
    return peer_requirements is not None


def submitter_is_finished(submission_uuid, peer_requirements):
    """
    Check whether the submitter has made the required number of assessments.

    If the requirements dict is None (because we're being updated
    asynchronously or when the workflow is first created),
    then automatically return False.

    Args:
        submission_uuid (str): The UUID of the submission being tracked.
        peer_requirements (dict): Dictionary with the key "must_grade" indicating
            the required number of submissions the student must grade.

    Returns:
        bool

    """
    if peer_requirements is None:
        return False

    try:
        workflow = PeerWorkflow.objects.get(submission_uuid=submission_uuid)
        if workflow.completed_at is not None:
            return True
        elif workflow.num_peers_graded() >= peer_requirements["must_grade"]:
            workflow.completed_at = timezone.now()
            workflow.save()
            return True
        return False
    except PeerWorkflow.DoesNotExist:
        return False
    except KeyError as ex:
        raise PeerAssessmentRequestError('Requirements dict must contain "must_grade" key') from ex


def get_graded_by_count(submission_uuid):
    """
    Retrieve the number of peer assessments the submitter has received.
    Returns None if no submission with this ID.
    """
    workflow = PeerWorkflow.get_by_submission_uuid(submission_uuid)
    if workflow is None:
        return None

    scored_items = workflow.graded_by.filter(
        assessment__submission_uuid=submission_uuid,
        assessment__score_type=PEER_TYPE
    )
    return scored_items.count()


def assessment_is_finished(submission_uuid, peer_requirements):
    """
    Check whether the submitter has received enough assessments
    to get a score.

    If the requirements dict is None (because we're being updated
    asynchronously or when the workflow is first created),
    then automatically return False.

    Args:
        submission_uuid (str): The UUID of the submission being tracked.
        peer_requirements (dict): Dictionary with the key "must_be_graded_by"
            indicating the required number of assessments the student
            must receive to get a score.

    Returns:

        bool
    """
    if not peer_requirements:
        return False

    count = get_graded_by_count(submission_uuid)
    if count is None:
        return False

    return count >= required_peer_grades(submission_uuid, peer_requirements)


def on_start(submission_uuid):
    """Create a new peer workflow for a student item and submission.

    Creates a unique peer workflow for a student item, associated with a
    submission.

    Args:
        submission_uuid (str): The submission associated with this workflow.

    Returns:
        None

    Raises:
        SubmissionError: There was an error retrieving the submission.
        PeerAssessmentInternalError: Raised when there is an internal error
            creating the Workflow.

    """
    try:
        with transaction.atomic():
            submission = sub_api.get_submission_and_student(submission_uuid)
            workflow, __ = PeerWorkflow.objects.get_or_create(
                student_id=submission['student_item']['student_id'],
                course_id=submission['student_item']['course_id'],
                item_id=submission['student_item']['item_id'],
                submission_uuid=submission_uuid
            )
            workflow.save()
    except IntegrityError:
        # If we get an integrity error, it means someone else has already
        # created a workflow for this submission, so we don't need to do anything.
        pass
    except DatabaseError as ex:
        error_message = (
            "An internal error occurred while creating a new peer "
            "workflow for submission {}"
            .format(submission_uuid)
        )
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def get_score(submission_uuid, peer_requirements):
    """
    Retrieve a score for a submission if requirements have been satisfied.

    Args:
        submission_uuid (str): The UUID of the submission.
        requirements (dict): Dictionary with the key "must_be_graded_by"
            indicating the required number of assessments the student
            must receive to get a score.

    Returns:
        A dictionary with the points earned, points possible, and
        contributing_assessments information, along with a None staff_id.

    """

    if peer_requirements is None:
        return None

    # User hasn't completed their own submission yet
    if not submitter_is_finished(submission_uuid, peer_requirements):
        return None

    workflow = PeerWorkflow.get_by_submission_uuid(submission_uuid)

    if workflow is None:
        return None

    # Retrieve the assessments in ascending order by score date,
    # because we want to use the *first* one(s) for the score.
    items = workflow.graded_by.filter(
        assessment__submission_uuid=submission_uuid,
        assessment__score_type=PEER_TYPE
    ).order_by('-assessment')

    # Check if enough peers have graded this submission
    if items.count() < required_peer_grades(submission_uuid, peer_requirements):
        return None

    # Unfortunately, we cannot use update() after taking a slice,
    # so we need to update the and save the items individually.
    # One might be tempted to first query for the first n assessments,
    # then select items that have those assessments.
    # However, this generates a SQL query with a LIMIT in a subquery,
    # which is not supported by some versions of MySQL.
    # Although this approach generates more database queries, the number is likely to
    # be relatively small (at least 1 and very likely less than 5).
    for scored_item in items[:peer_requirements["must_be_graded_by"]]:
        scored_item.scored = True
        scored_item.save()
    assessments = [item.assessment for item in items]

    return {
        "points_earned": sum(
            get_assessment_median_scores(submission_uuid).values()
        ),
        "points_possible": assessments[0].points_possible,
        "contributing_assessments": [assessment.id for assessment in assessments],
        "staff_id": None,
    }


def create_assessment(
        scorer_submission_uuid,
        scorer_id,
        options_selected,
        criterion_feedback,
        overall_feedback,
        rubric_dict,
        num_required_grades,
        scored_at=None
):
    # pylint: disable=unicode-format-string
    """
    Creates an assessment on the given submission.

    Assessments are created based on feedback associated with a particular
    rubric.

    Args:
        scorer_submission_uuid (str): The submission uuid for the Scorer's
            workflow. The submission being assessed can be determined via the
            peer workflow of the grading student.
        scorer_id (str): The user ID for the user giving this assessment. This
            is required to create an assessment on a submission.
        options_selected (dict): Dictionary mapping criterion names to the
            option names the user selected for that criterion.
        criterion_feedback (dict): Dictionary mapping criterion names to the
            free-form text feedback the user gave for the criterion.
            Since criterion feedback is optional, some criteria may not appear
            in the dictionary.
        overall_feedback (unicode): Free-form text feedback on the submission overall.
        num_required_grades (int): The required number of assessments a
            submission requires before it is completed. If this number of
            assessments is reached, the grading_completed_at timestamp is set
            for the Workflow.

    Keyword Args:
        scored_at (datetime): Optional argument to override the time in which
            the assessment took place. If not specified, scored_at is set to
            now.

    Returns:
        dict: the Assessment model, serialized as a dict.

    Raises:
        PeerAssessmentRequestError: Raised when the submission_id is invalid, or
            the assessment_dict does not contain the required values to create
            an assessment.
        PeerAssessmentInternalError: Raised when there is an internal error
            while creating a new assessment.

    Examples:
        >>> options_selected = {"clarity": "Very clear", "precision": "Somewhat precise"}
        >>> criterion_feedback = {"clarity": "I thought this essay was very clear."}
        >>> feedback = "Your submission was thrilling."
        >>> create_assessment("1", "Tim", options_selected, criterion_feedback, feedback, rubric_dict)
    """
    try:
        # Retrieve workflow information
        scorer_workflow = PeerWorkflow.objects.get(submission_uuid=scorer_submission_uuid)
        peer_workflow_item = scorer_workflow.find_active_assessments()
        if peer_workflow_item is None:
            message = (
                "There are no open assessments associated with the scorer's "
                "submission UUID {}."
            ).format(scorer_submission_uuid)
            logger.warning(message)
            raise PeerAssessmentWorkflowError(message)
        peer_submission_uuid = peer_workflow_item.submission_uuid

        assessment = _complete_assessment(
            rubric_dict,
            scorer_id,
            peer_submission_uuid,
            options_selected,
            criterion_feedback,
            scorer_workflow,
            overall_feedback,
            num_required_grades,
            scored_at
        )

        _log_assessment(assessment, scorer_workflow)
        return full_assessment_dict(assessment)
    except PeerWorkflow.DoesNotExist as ex:
        message = (
            "There is no Peer Workflow associated with the given "
            "submission UUID {}."
        ).format(scorer_submission_uuid)
        logger.exception(message)
        raise PeerAssessmentWorkflowError(message) from ex
    except InvalidRubric as ex:
        msg = "The rubric definition is not valid."
        logger.exception(msg)
        raise PeerAssessmentRequestError(msg) from ex
    except InvalidRubricSelection as ex:
        msg = "Invalid options were selected in the rubric."
        logger.warning(msg, exc_info=True)
        raise PeerAssessmentRequestError(msg) from ex
    except DatabaseError as ex:
        error_message = (
            "An error occurred while creating an assessment by the scorer with this ID: {}"
        ).format(scorer_id)
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


@transaction.atomic
def _complete_assessment(
        rubric_dict,
        scorer_id,
        peer_submission_uuid,
        options_selected,
        criterion_feedback,
        scorer_workflow,
        overall_feedback,
        num_required_grades,
        scored_at
):
    """
    Internal function for atomic assessment creation. Creates a peer assessment
    and closes the associated peer workflow item in a single transaction.

    Args:
        rubric_dict (dict): The rubric model associated with this assessment
        scorer_id (str): The user ID for the user giving this assessment. This
            is required to create an assessment on a submission.
        peer_submission_uuid (str): The submission uuid for the submission being
            assessed.
        options_selected (dict): Dictionary mapping criterion names to the
            option names the user selected for that criterion.
        criterion_feedback (dict): Dictionary mapping criterion names to the
            free-form text feedback the user gave for the criterion.
            Since criterion feedback is optional, some criteria may not appear
            in the dictionary.
        scorer_workflow (PeerWorkflow): The PeerWorkflow associated with the
            scorer. Updates the workflow item associated with this assessment.
        overall_feedback (unicode): Free-form text feedback on the submission overall.
        num_required_grades (int): The required number of assessments a
            submission requires before it is completed. If this number of
            assessments is reached, the grading_completed_at timestamp is set
            for the Workflow.
        scored_at (datetime): Optional argument to override the time in which
            the assessment took place. If not specified, scored_at is set to
            now.

    Returns:
        The Assessment model

    """
    # Get or create the rubric
    rubric = rubric_from_dict(rubric_dict)

    # Create the peer assessment
    assessment = Assessment.create(
        rubric,
        scorer_id,
        peer_submission_uuid,
        PEER_TYPE,
        scored_at=scored_at,
        feedback=overall_feedback
    )

    # Create assessment parts for each criterion in the rubric
    # This will raise an `InvalidRubricSelection` if the selected options do not
    # match the rubric.
    AssessmentPart.create_from_option_names(assessment, options_selected, feedback=criterion_feedback)

    # Close the active assessment
    scorer_workflow.close_active_assessment(peer_submission_uuid, assessment, num_required_grades)
    return assessment


def get_rubric_max_scores(submission_uuid):
    """Gets the maximum possible value for each criterion option

    Iterates over the rubric used to grade the given submission, and creates a
    dictionary of maximum possible values.

    Args:
        submission_uuid: The submission to get the associated rubric max scores.
    Returns:
        A dictionary of max scores for this rubric's criteria options. Returns
            None if no assessments are found for this submission.
    Raises:
        PeerAssessmentInternalError: Raised when there is an error retrieving
            the submission, or its associated rubric.
    """
    try:
        assessments = list(
            Assessment.objects.filter(
                submission_uuid=submission_uuid
            ).order_by("-scored_at", "-id").select_related("rubric")[:1]
        )
        if not assessments:
            return None

        assessment = assessments[0]
        rubric_dict = RubricSerializer.serialized_from_cache(assessment.rubric)
        return {
            criterion["name"]: criterion["points_possible"]
            for criterion in rubric_dict["criteria"]
        }
    except DatabaseError as ex:
        error_message = (
            "Error getting rubric options max scores for submission uuid {uuid}"
        ).format(uuid=submission_uuid)
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def get_assessment_median_scores(submission_uuid):
    """Get the median score for each rubric criterion

    For a given assessment, collect the median score for each criterion on the
    rubric. This set can be used to determine the overall score, as well as each
    part of the individual rubric scores.

    If there is a true median score, it is returned. If there are two median
    values, the average of those two values is returned, rounded up to the
    greatest integer value.

    Args:
        submission_uuid (str): The submission uuid is used to get the
            assessments used to score this submission, and generate the
            appropriate median score.

    Returns:
        dict: A dictionary of rubric criterion names,
        with a median score of the peer assessments.

    Raises:
        PeerAssessmentInternalError: If any error occurs while retrieving
            information to form the median scores, an error is raised.
    """
    try:
        workflow = PeerWorkflow.objects.get(submission_uuid=submission_uuid)
        items = workflow.graded_by.filter(scored=True)
        assessments = [item.assessment for item in items]
        scores = Assessment.scores_by_criterion(assessments)
        return Assessment.get_median_score_dict(scores)
    except PeerWorkflow.DoesNotExist:
        return {}
    except DatabaseError as ex:
        error_message = (
            "Error getting assessment median scores for submission {uuid}"
        ).format(uuid=submission_uuid)
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def has_finished_required_evaluating(submission_uuid, required_assessments):
    """Check if a student still needs to evaluate more submissions

    Per the contract of the peer assessment workflow, a student must evaluate a
    number of peers before receiving feedback on their submission.

    Args:
        submission_uuid (str): The submission UUID is required to determine if
            the associated student has completed enough assessments. This
            argument is required.
        required_assessments (int): The number of assessments a student has to
            submit before receiving the feedback on their submission. This is a
            required argument.

    Returns:
        tuple: True if the student has evaluated enough peer submissions to move
            through the peer assessment workflow. False if the student needs to
            evaluate more peer submissions. The second value is the count of
            assessments completed.

    Raises:
        PeerAssessmentRequestError: Raised when the submission UUID is invalid,
            or the required_assessments is not a positive integer.
        PeerAssessmentInternalError: Raised when there is an internal error
            while evaluating this workflow rule.

    Examples:
        >>> has_finished_required_evaluating("abc123", 3)
        True, 3

    """
    workflow = PeerWorkflow.get_by_submission_uuid(submission_uuid)
    done = False
    peers_graded = 0
    if workflow:
        peers_graded = workflow.num_peers_graded()
        done = (peers_graded >= required_assessments)
    return done, peers_graded


def get_assessments(submission_uuid, limit=None):
    """Retrieve the assessments for a submission.

    Retrieves all the assessments for a submissions. This API returns related
    feedback without making any assumptions about grading. Any outstanding
    assessments associated with this submission will not be returned.

    Args:
        submission_uuid (str): The submission all the requested assessments are
            associated with. Required.

    Keyword Arguments:
        limit (int): Limit the returned assessments. If None, returns all.


    Returns:
        list: A list of dictionaries, where each dictionary represents a
            separate assessment. Each assessment contains points earned, points
            possible, time scored, scorer id, score type, and feedback.


    Raises:
        PeerAssessmentRequestError: Raised when the submission_id is invalid.
        PeerAssessmentInternalError: Raised when there is an internal error
            while retrieving the assessments associated with this submission.

    Examples:
        >>> get_assessments("1", limit=2)
        [
            {
                'points_earned': 6,
                'points_possible': 12,
                'scored_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 649284 tzinfo=<UTC>),
                'scorer': u"Tim",
                'feedback': u'Your submission was thrilling.'
            },
            {
                'points_earned': 11,
                'points_possible': 12,
                'scored_at': datetime.datetime(2014, 1, 31, 14, 10, 17, 544214 tzinfo=<UTC>),
                'scorer': u"Bob",
                'feedback': u'Great submission.'
            }
        ]

    """
    try:
        assessments = Assessment.objects.filter(
            submission_uuid=submission_uuid,
            score_type=PEER_TYPE
        )[:limit]
        return serialize_assessments(assessments)
    except DatabaseError as ex:
        error_message = (
            "Error getting assessments for submission {uuid}"
        ).format(uuid=submission_uuid)
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def get_submitted_assessments(submission_uuid, limit=None):
    """Retrieve the assessments created by the given submission's author.

    Retrieves all the assessments created by the given submission's author. This
    API returns related feedback without making any assumptions about grading.
    Any outstanding assessments associated with this submission will not be
    returned.

    Args:
        submission_uuid (str): The submission of the student whose assessments
        we are requesting. Required.

    Keyword Arguments:
        limit (int): Limit the returned assessments. If None, returns all.

    Returns:
        list(dict): A list of dictionaries, where each dictionary represents a
            separate assessment. Each assessment contains points earned, points
            possible, time scored, scorer id, score type, and feedback. If no
            workflow is found associated with the given submission_uuid, returns
            an empty list.

    Raises:
        PeerAssessmentRequestError: Raised when the submission_id is invalid.
        PeerAssessmentInternalError: Raised when there is an internal error
            while retrieving the assessments associated with this submission.

    Examples:
        >>> get_submitted_assessments("1", limit=2)
        [
            {
                'points_earned': 6,
                'points_possible': 12,
                'scored_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 649284 tzinfo=<UTC>),
                'scorer': u"Tim",
                'feedback': u'Your submission was thrilling.'
            },
            {
                'points_earned': 11,
                'points_possible': 12,
                'scored_at': datetime.datetime(2014, 1, 31, 14, 10, 17, 544214 tzinfo=<UTC>),
                'scorer': u"Tim",
                'feedback': u'Great submission.'
            }
        ]

    """
    try:
        # If no workflow is found associated with the uuid, this returns None,
        # and an empty set of assessments will be returned.
        workflow = PeerWorkflow.get_by_submission_uuid(submission_uuid)
        items = PeerWorkflowItem.objects.filter(
            scorer=workflow,
            assessment__isnull=False
        )
        assessments = Assessment.objects.filter(
            pk__in=[item.assessment.pk for item in items])[:limit]
        return serialize_assessments(assessments)
    except DatabaseError as ex:
        error_message = (
            "Couldn't retrieve the assessments completed by the student with submission {uuid}"
        ).format(uuid=submission_uuid)
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def get_submission_to_assess(submission_uuid, graded_by):
    """Get a submission to peer evaluate.

    Retrieves a submission for assessment for the given student. This will
    not return a submission submitted by the requesting scorer. Submissions are
    returned based on how many assessments are still required, and if there are
    peers actively assessing a particular submission. If there are no
    submissions requiring assessment, a submission may be returned that will be
    'over graded', and the assessment will not be counted towards the overall
    grade.

    Args:
        submission_uuid (str): The submission UUID from the student
            requesting a submission for assessment. This is used to explicitly
            avoid giving the student their own submission, and determines the
            associated Peer Workflow.
        graded_by (int): The number of assessments a submission
            requires before it has completed the peer assessment process.

    Returns:
        dict: A peer submission for assessment. This contains a 'student_item',
            'attempt_number', 'submitted_at', 'created_at', and 'answer' field to be
            used for assessment.

    Raises:
        PeerAssessmentRequestError: Raised when the request parameters are
            invalid for the request.
        PeerAssessmentInternalError: Raised when there is an internal error
            retrieving peer workflow information.
        PeerAssessmentWorkflowError: Raised when an error occurs because this
            function, or the student item, is not in the proper workflow state
            to retrieve a peer submission.

    Examples:
        >>> get_submission_to_assess("abc123", 3)
        {
            'student_item': 2,
            'attempt_number': 1,
            'submitted_at': datetime.datetime(2014, 1, 29, 23, 14, 52, 649284, tzinfo=<UTC>),
            'created_at': datetime.datetime(2014, 1, 29, 17, 14, 52, 668850, tzinfo=<UTC>),
            'answer': u'The answer is 42.'
        }

    """
    workflow = PeerWorkflow.get_by_submission_uuid(submission_uuid)

    if not workflow:
        raise PeerAssessmentWorkflowError(
            "A Peer Assessment Workflow does not exist for the student "
            "with submission UUID {}".format(submission_uuid)
        )

    if workflow.is_cancelled:
        return None

    open_item = workflow.find_active_assessments()
    peer_submission_uuid = open_item.submission_uuid if open_item else None
    # If there is an active assessment for this user, get that submission,
    # otherwise, get the first assessment for review, otherwise,
    # get the first submission available for over grading ("over-grading").
    if peer_submission_uuid is None:
        peer_submission_uuid = workflow.get_submission_for_review(graded_by)
    if peer_submission_uuid is None:
        peer_submission_uuid = workflow.get_submission_for_over_grading()
    if peer_submission_uuid:
        try:
            submission_data = sub_api.get_submission(peer_submission_uuid)
            PeerWorkflow.create_item(workflow, peer_submission_uuid)
            _log_workflow(peer_submission_uuid, workflow)
            return submission_data
        except sub_api.SubmissionNotFoundError as ex:
            error_message = "Could not find a submission with the uuid %s for student %s in the peer workflow."
            error_meesage_args = (peer_submission_uuid, workflow.student_id)
            logger.exception(error_message, error_meesage_args[0], error_meesage_args[1])
            raise PeerAssessmentWorkflowError(error_message % error_meesage_args) from ex
    else:
        logger.info(
            "No submission found for %s to assess (%s, %s)",
            workflow.student_id,
            workflow.course_id,
            workflow.item_id
        )
        return None


def create_peer_workflow(submission_uuid):
    """Create a new peer workflow for a student item and submission.

    Creates a unique peer workflow for a student item, associated with a
    submission.

    Args:
        submission_uuid (str): The submission associated with this workflow.

    Returns:
        None

    Raises:
        SubmissionError: There was an error retrieving the submission.
        PeerAssessmentInternalError: Raised when there is an internal error
            creating the Workflow.

    Examples:
        >>> create_peer_workflow("1")

    """
    try:
        with transaction.atomic():
            submission = sub_api.get_submission_and_student(submission_uuid)
            workflow, __ = PeerWorkflow.objects.get_or_create(
                student_id=submission['student_item']['student_id'],
                course_id=submission['student_item']['course_id'],
                item_id=submission['student_item']['item_id'],
                submission_uuid=submission_uuid
            )
            workflow.save()
    except IntegrityError:
        # If we get an integrity error, it means someone else has already
        # created a workflow for this submission, so we don't need to do anything.
        pass
    except DatabaseError as ex:
        error_message = (
            "An internal error occurred while creating a new peer "
            "workflow for submission {}"
        ).format(submission_uuid)
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def create_peer_workflow_item(scorer_submission_uuid, submission_uuid):
    """
    Begin peer-assessing a particular submission.
    Note that this does NOT pick the submission from the prioritized list of available submissions.
    Mainly useful for testing.

    Args:
        scorer_submission_uuid (str): The ID of the scoring student.
        submission_uuid (str): The unique identifier of the submission being scored

    Returns:
        None

    Raises:
        PeerAssessmentWorkflowError: Could not find the workflow for the student.
        PeerAssessmentInternalError: Could not create the peer workflow item.
    """
    workflow = PeerWorkflow.get_by_submission_uuid(scorer_submission_uuid)
    PeerWorkflow.create_item(workflow, submission_uuid)


def get_assessment_feedback(submission_uuid):
    """
    Retrieve a feedback on an assessment.

    Args:
        submission_uuid: The submission we want to retrieve assessment feedback for.

    Returns:
        dict or None

    Raises:
        PeerAssessmentInternalError: Error occurred while retrieving the feedback.
    """
    try:
        feedback = AssessmentFeedback.objects.get(
            submission_uuid=submission_uuid
        )
        return AssessmentFeedbackSerializer(feedback).data
    except AssessmentFeedback.DoesNotExist:
        return None
    except DatabaseError as ex:
        error_message = (
            "An error occurred retrieving assessment feedback for {}."
            .format(submission_uuid)
        )
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def set_assessment_feedback(feedback_dict):
    """
    Set a feedback object for an assessment to have some new values.

    Sets or updates the assessment feedback with the given values in the dict.

    Args:
        feedback_dict (dict): A dictionary of all the values to update or create
            a new assessment feedback.

    Returns:
        None

    Raises:
        PeerAssessmentRequestError
        PeerAssessmentInternalError
    """
    submission_uuid = feedback_dict.get('submission_uuid')
    feedback_text = feedback_dict.get('feedback_text')
    selected_options = feedback_dict.get('options', [])

    if feedback_text and len(feedback_text) > AssessmentFeedback.MAXSIZE:
        error_message = "Assessment feedback too large."
        raise PeerAssessmentRequestError(error_message)

    try:
        # Get or create the assessment model for this submission
        # If we receive an integrity error, assume that someone else is trying to create
        # another feedback model for this submission, and raise an exception.
        if submission_uuid:
            feedback, created = AssessmentFeedback.objects.get_or_create(submission_uuid=submission_uuid)
        else:
            error_message = "An error occurred creating assessment feedback: bad or missing submission_uuid."
            logger.error(error_message)
            raise PeerAssessmentRequestError(error_message)

        # Update the feedback text
        if feedback_text is not None:
            feedback.feedback_text = feedback_text

        # Save the feedback model.  We need to do this before setting m2m relations.
        if created or feedback_text is not None:
            feedback.save()

        # Associate the feedback with selected options
        feedback.add_options(selected_options)

        # Associate the feedback with scored assessments
        assessments = PeerWorkflowItem.get_scored_assessments(submission_uuid)
        feedback.assessments.add(*assessments)
    except DatabaseError as ex:
        msg = f"Error occurred while creating or updating feedback on assessment: {feedback_dict}"
        logger.exception(msg)
        raise PeerAssessmentInternalError(msg) from ex


def _log_assessment(assessment, scorer_workflow):
    """
    Log the creation of a peer assessment.

    Args:
        assessment (Assessment): The assessment model that was created.
        scorer_workflow (dict): A dictionary representation of the Workflow
            belonging to the scorer of this assessment.

    Returns:
        None

    """
    logger.info(
        "Created peer-assessment %s for submission %s, course %s, item %s with rubric %s; scored by %s",
        assessment.id,
        assessment.submission_uuid,
        scorer_workflow.course_id,
        scorer_workflow.item_id,
        assessment.rubric.content_hash,
        scorer_workflow.student_id,
    )


def _log_workflow(submission_uuid, workflow):
    """
    Log the creation of a peer-assessment workflow.

    Args:
        submission_uuid (str): The UUID of the submission being assessed.
        workflow (PeerWorkflow): The Peer Workflow of the student making the
            assessment.
    """
    logger.info(
        "Retrieved submission %s (%s, %s) to be assessed by %s",
        submission_uuid,
        workflow.course_id,
        workflow.item_id,
        workflow.student_id,
    )


def is_workflow_cancelled(submission_uuid):
    """
    Check if workflow submission is cancelled.

    Args:
        submission_uuid (str): The UUID of the workflow's submission.

    Returns:
        True/False
    """
    if submission_uuid is None:
        return False
    try:
        workflow = PeerWorkflow.get_by_submission_uuid(submission_uuid)
        return workflow.is_cancelled if workflow else False
    except PeerAssessmentWorkflowError:
        return False


def on_cancel(submission_uuid):
    """Cancel the peer workflow for submission.

    Sets the cancelled_at field in peer workflow.

    Args:
        submission_uuid (str): The submission UUID associated with this workflow.

    Returns:
        None

    """
    try:
        workflow = PeerWorkflow.get_by_submission_uuid(submission_uuid)
        if workflow:
            workflow.cancelled_at = timezone.now()
            workflow.save()
    except (PeerAssessmentWorkflowError, DatabaseError) as ex:
        error_message = (
            "An internal error occurred while cancelling the peer"
            "workflow for submission {}"
            .format(submission_uuid)
        )
        logger.exception(error_message)
        raise PeerAssessmentInternalError(error_message) from ex


def get_waiting_step_details(
    course_id,
    item_id,
    submission_uuids,
    must_be_graded_by
):
    """
    Proxy method to `get_waiting_step_details` model method.
    Retrieves information about users in the waiting step (waiting for peer reviews).

    Args:
        course_id (str): The course that this problem belongs to.
        item_id (str): The student_item (problem) that we want to know statistics about.
        submission_uuids (list): A list of submission UUIDs to filter the results for,
                                    if None is given, this will return all students which
                                    the peer step is not complete.
        must_be_graded_by (int): number of required peer reviews for this problem.

    Returns:
        dict: a dictionary that contains information about students in the waiting step.
              The dictionary includes the following information: `student_id`, `created_at` (
              timestamp of when the step was created), `graded` (how many peers the student
              graded) and `graded_by` (how many peers graded this student).
    """
    return PeerWorkflow.get_waiting_step_details(
        course_id,
        item_id,
        submission_uuids,
        must_be_graded_by
    )



"""
Django admin models for staffgrader
"""

from django.contrib import admin
from openassessment.staffgrader.models import SubmissionGradingLock


class SubmissionGradingLockAdmin(admin.ModelAdmin):
    """
    Django admin model for SubmissionGradingLock.
    """
    list_display = ('id', 'submission_uuid', 'owner_id', 'created_at', 'is_active')
    readonly_fields = ('is_active',)
    search_fields = ('submission_uuid',)

    # This allows us to have the nice boolean check/x icons in the list rather than "True"/"False"
    def is_active(self, lock):
        return lock.is_active
    is_active.boolean = True


admin.site.register(SubmissionGradingLock, SubmissionGradingLockAdmin)



"""
The File Upload application is designed to allow the management of files
associated with submissions. This can be used to upload new files, manage
URLs of existing files, and delete files.
"""

from collections import namedtuple
import json
import logging

from django.db import IntegrityError
from django.utils.functional import cached_property

from openassessment.assessment.models.base import SharedFileUpload
from openassessment.fileupload.exceptions import FileUploadError

from . import backends


logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

KEY_SEPARATOR = '/'


def get_upload_url(key, content_type):
    """
    Returns a url (absolute or relative, depending on the endpoint) which can be used to upload a file to.
    """
    return backends.get_backend().get_upload_url(key, content_type)


def get_download_url(key):
    """
    Returns the url at which the file that corresponds to the key can be downloaded.
    """
    url = backends.get_backend().get_download_url(key)
    if not url:
        logger.warning('FileUploadError: Could not retrieve URL for key %s', key)
    return url


def remove_file(key):
    """
    Remove file from the storage
    """
    return backends.get_backend().remove_file(key)


def get_student_file_key(student_item_dict, index=0):
    """
    Args:
        student_item_dict: A dictionary containing keys ('student_id', 'course_id', 'item_id').
        index (int, optional): The index of a file.
    """
    key_template = KEY_SEPARATOR.join(('{student_id}', '{course_id}', '{item_id}'))
    index = int(index)
    if index > 0:
        key_template += KEY_SEPARATOR + '{index}'
    return key_template.format(index=index, **student_item_dict)


def can_delete_file(current_user_id, teams_enabled, key, team_id=None, shared_file=None):
    """
    A user is allowed to delete any file they own if this is not a team-enabled response.
    If the response is team-enabled, a user, who is a member of a team,
    is allowed to delete a file they own as long as they are still
    a member of the team with which the file has been shared.

    params:
      current_user_id (string): The anonymous id of the current user in an ORA block.
      teams_enabled (boolean): Indicates if teams are enabled for an ORA block.
      key (string): The key of the file to check if we can delete.
      team_id (string): The id of the team of the user who may be able to delete the file.
      shared_file (SharedFileUpload): Optional. A SharedFileUpload object corresponding to the given
      key.  It's useful to pass this in if you've already fetched all of the SharedFileUpload records
      for a given item/team.

    raises:
        SharedFileUpload.DoesNotExist If teams are enabled, a team_id is provided,
        and no SharedFileUpload corresponding to the file key exists.
    returns:
        Boolean indicating if the file with the given key can be deleted by the current user.
    """
    if not teams_enabled:
        return True

    if not shared_file:
        try:
            shared_file = SharedFileUpload.by_key(key)
        except SharedFileUpload.DoesNotExist:
            logger.info('While checking ORA file-deletion ability, could not find file with key: %s', key)
            return True

    if shared_file.owner_id != current_user_id:
        return False

    if shared_file.team_id != team_id:
        return False

    # If we've made it this far, the current user has a team, and it's the same
    # team that the file is shared with, so let them (as the file's owner) delete it.
    return True


def delete_shared_files_for_team(course_id, item_id, team_id):
    """
    Delete shared files for a team for this block
    """
    uploads = SharedFileUpload.by_team_course_item(team_id, course_id, item_id)

    for upload in uploads:
        remove_file(upload.file_key)
        upload.delete()


def _safe_load_json_list(field, log_error=False):
    """
    Tries to load JSON-ified string,
    returns an empty list if we try to load some non-JSON-encoded string.
    """
    try:
        return json.loads(field)
    except ValueError:
        if log_error:
            logger.exception(
                "URLWorkaround: Safe Load failed for data field:%s with type:%s", field, type(field)
            )
        return []


class FileUpload:
    """
    A layer of abstraction over the various components of file
    data stored as ORA XBlock user-scoped fields.
    """
    def __init__(self, name=None, description=None, size=None, index=0, descriptionless=False, **student_item_dict):
        """
        Args:
            name (str): The name of a file.
            description (str): The student-provided description of a file.
            size (int): The size, in bytes, of a file.
            index (int): The position of a file relative to all other uploaded files for a given user.
            descriptionless (bool): True if this file exists but has no description, name, or size.
                                    False (default) otherwise.
            student_item_dict (dict): Contains the student_id, course_id, and item_id, i.e. the "student item"
                                      triple associated with this file upload.
        """
        self.name = name
        self.description = description
        self.size = size
        self.index = index
        self.student_id = student_item_dict.get('student_id')
        self.course_id = student_item_dict.get('course_id')
        self.item_id = student_item_dict.get('item_id')
        self.descriptionless = descriptionless

    @property
    def exists(self):
        return (self.description is not None) or self.descriptionless

    @property
    def download_url(self):
        """
        Returns the url at which the file that corresponds to the key
        can be downloaded if exists.
        """
        if self.exists:
            try:
                return get_download_url(self.key)
            except FileUploadError as exc:
                logger.exception(
                    'FileUploadError: URL retrieval failed for key %s with error %s',
                    self.key,
                    exc,
                    exc_info=True,
                )
                return ''
        return None

    @property
    def key(self):
        """
        Simple utility method to generate a common file upload key based on
        the student item.

        Returns:
            A string representation of the key.
        """
        student_item_dict = {
            'student_id': self.student_id,
            'course_id': self.course_id,
            'item_id': self.item_id,
        }
        return get_student_file_key(student_item_dict, index=self.index)

    def _to_dict(self):
        """
        Returns:
            A dictionary representation of the FileUpload.
        """
        attrs = ('description', 'name', 'size', 'course_id', 'student_id', 'item_id', 'descriptionless')
        return {
            key: getattr(self, key, None) for key in attrs
        }

    def __eq__(self, other):
        """
        Returns:
            True if self's dict representation equals other's dict representation,
            False otherwise.
        """
        return self._to_dict() == other._to_dict()  # pylint: disable=protected-access

    def __hash__(self):
        """
        Returns a hash of the FileUpload's dict representation
        """
        return hash(self._to_dict())


FileDescriptor = namedtuple('FileDescriptor', ['download_url', 'description', 'name', 'size', 'show_delete_button'])
TeamFileDescriptor = namedtuple('TeamFileDescriptor', ['download_url', 'description', 'name', 'size', 'uploaded_by'])


class FileUploadManager:
    """
    Manages the CRUD operations of file uploads
    that take place in the context of an OpenAssessmentBlock.

    e.g. inside an XBlock:
    self.upload_manager = FileUploadManager(self)
    for file_upload in self.upload_manager.get_uploads():
        log.info(file_upload.download_url)

    new_uploads = [
        {'name': 'file-1.jpg', 'description': 'File 1', 'size': 1024},
        {'name': 'file-2.jpg', 'description': 'File 2', 'size': 2048},
    ]
    self.upload_manager.append_uploads(*new_uploads)

    self.upload_manager.delete_upload(index=0)
    """
    def __init__(self, openassessment_xblock):
        self.block = openassessment_xblock
        self.shared_uploads_for_team_by_key_cache = {}

    @cached_property
    def student_item_dict(self):
        """ Returns a dict containing 'student_id', 'course_id', and 'item_id'. """
        return self.block.get_student_item_dict()

    def get_uploads(self, team_id=None, include_deleted=False):
        """
        Returns:
            A list of FileUpload objects associated with an instance of an Open Assessment Block.
            This will include FileUpload objects corresponding to existing files.
            It does **not** include entries for files that have been deleted.
        """
        descriptions, names, sizes = self._get_metadata_from_block()
        user_uploads = self._file_uploads_from_list_fields(descriptions, names, sizes, include_deleted=include_deleted)

        if self.block.is_team_assignment():
            return self._uploads_shared_with_team_by_current_user(user_uploads, team_id)

        return user_uploads

    def get_team_uploads(self, team_id=None):
        if self.block.is_team_assignment():
            return self._uploads_owned_by_teammates(team_id)
        return []

    def _uploads_shared_with_team_by_current_user(self, user_uploads, team_id):
        """
        Helper function that filters a given list of ``user_uploads``
        down to those ``FileUploads`` that are owned by the current user
        **and** the given team
        """
        jointly_owned_uploads = []

        for upload in user_uploads:
            shared_upload = self.shared_uploads_for_student_by_key.get(upload.key)
            if shared_upload and (shared_upload.team_id == team_id):
                jointly_owned_uploads.append(upload)
            elif not upload.exists:
                # we should return entries for deleted files, here,
                # to uphold the invariant around file indices.
                jointly_owned_uploads.append(upload)

        return jointly_owned_uploads

    def _uploads_owned_by_teammates(self, team_id):
        """
        Returns a list of FileUpload objects owned by other members of the given team.
        Does not include FileUploads of the current user.
        """
        shared_uploads_from_other_users = sorted(
            [
                shared_upload
                for shared_upload in self.shared_uploads_for_team_by_key(team_id).values()
                if shared_upload.owner_id != self.student_item_dict['student_id']
            ],
            key=lambda upload: upload.file_key,
        )

        return [
            FileUpload(
                name=shared_upload.name,
                description=shared_upload.description,
                size=shared_upload.size,
                student_id=shared_upload.owner_id,
                course_id=shared_upload.course_id,
                item_id=shared_upload.item_id,
                index=shared_upload.index,
            ) for shared_upload in shared_uploads_from_other_users
        ]

    def file_descriptors(self, team_id=None, include_deleted=False):
        """
        Used in the response template context to provide file information
        (file URL, description, name, show_delete_button) for each uploaded
        file in this block.

        If self.block is team-enabled, this will return only entries for files
        that have been shared with the specified team
        """

        descriptors = []

        for upload in self.get_uploads(team_id=team_id, include_deleted=include_deleted):
            show_delete_button = bool(upload.exists)

            if upload.exists and self.block.is_team_assignment():
                shared_upload = self.shared_uploads_for_team_by_key(team_id)[upload.key]
                show_delete_button = can_delete_file(
                    self.student_item_dict['student_id'],
                    self.block.is_team_assignment(),
                    upload.key,
                    team_id=team_id,
                    shared_file=shared_upload,
                )

            descriptors.append(FileDescriptor(
                download_url=upload.download_url,
                description=upload.description,
                name=upload.name,
                size=upload.size,
                show_delete_button=show_delete_button,
            )._asdict())

        return descriptors

    def team_file_descriptors(self, team_id=None):
        """
        Returns the list of TeamFileDescriptors owned by other team members
        shown to a user when self.block is a team assignment.
        """
        return [
            TeamFileDescriptor(
                download_url=upload.download_url,
                description=upload.description,
                name=upload.name,
                size=upload.size,
                uploaded_by=self.block.get_username(upload.student_id)
            )._asdict()
            for upload in self.get_team_uploads(team_id=team_id)
        ]

    @cached_property
    def shared_uploads_for_student_by_key(self):
        """
        Returns **and caches** all of the SharedFileUpload records
        for this student/course/item.
        """
        shared_uploads = SharedFileUpload.by_student_course_item(**self.student_item_dict)
        return {shared_upload.file_key: shared_upload for shared_upload in shared_uploads}

    def shared_uploads_for_team_by_key(self, team_id):
        """
        Returns **and caches** all of the SharedFileUpload records
        for this student/course/item and team.

        Realistically, only one team_id will ever be requested, but this is a simple enough pattern
        """
        if team_id not in self.shared_uploads_for_team_by_key_cache:
            shared_uploads = SharedFileUpload.by_team_course_item(
                team_id=team_id,
                course_id=self.student_item_dict['course_id'],
                item_id=self.student_item_dict['item_id'],
            )
            shared_uploads_for_team_by_key = {
                shared_upload.file_key: shared_upload for shared_upload in shared_uploads
            }
            self.shared_uploads_for_team_by_key_cache[team_id] = shared_uploads_for_team_by_key
        return self.shared_uploads_for_team_by_key_cache[team_id]

    def invalidate_cached_shared_file_dicts(self):
        """
        Invalidates SharedFileUpload records that we have cached.
        """
        if hasattr(self, 'shared_uploads_for_student_by_key'):
            del self.shared_uploads_for_student_by_key

        self.shared_uploads_for_team_by_key_cache = {}

    def append_uploads(self, *new_uploads):
        """
        Given lists of new file metadata, write the new metadata to our stored file metadata fields

        Args:
            descriptions_to_add: a list of file descriptions
            names_to_add: a list of file names
            sizes_to_add: a list of file sizes as integers

        Returns: newly updated file metadata fields
        """
        required_keys = ('description', 'name', 'size')
        try:
            (
                descriptions_to_add,
                names_to_add,
                sizes_to_add,
            ) = self._dicts_to_key_lists(new_uploads, required_keys)
        except FileUploadError as exc:
            logging.exception(
                "FileUploadError: Metadata save for %s failed with error %s",
                exc,
                new_uploads
            )
            raise

        existing_file_descriptions, existing_file_names, existing_file_sizes = self._get_metadata_from_block()

        new_descriptions = existing_file_descriptions + descriptions_to_add
        self._set_file_descriptions(new_descriptions)

        new_names = existing_file_names + names_to_add
        self._set_file_names(new_names)

        new_sizes = existing_file_sizes + sizes_to_add
        self._set_file_sizes(new_sizes)

        new_file_uploads = self._file_uploads_from_list_fields(new_descriptions, new_names, new_sizes)

        if self.block.is_team_assignment() and self.block.has_team():
            existing_file_upload_key_set = {
                fileupload.key for fileupload in
                self._file_uploads_from_list_fields(
                    existing_file_descriptions,
                    existing_file_names,
                    existing_file_sizes
                )
            }

            for new_file_upload in new_file_uploads:
                if new_file_upload.key not in existing_file_upload_key_set:
                    self.create_shared_upload(new_file_upload)

        self.invalidate_cached_shared_file_dicts()
        return new_file_uploads

    def create_shared_upload(self, fileupload):
        try:
            SharedFileUpload.objects.create(
                team_id=self.block.team.team_id,
                owner_id=fileupload.student_id,
                course_id=fileupload.course_id,
                item_id=fileupload.item_id,
                file_key=fileupload.key,
                description=fileupload.description,
                size=fileupload.size,
                name=fileupload.name,
            )
        except IntegrityError as e:
            logger.error("Unable to create shared upload. %s", str(e))
            raise e

    def get_file_key(self, index):
        return get_student_file_key(self.student_item_dict, index)

    def delete_upload(self, index):
        """
        Given a file index to remove, null out its metadata in our stored file metadata fields.
        This will also delete any ``SharedFileUpload`` records associated with the file's key
        (if the file has been shared with a team).

        Args:
            index (integer): file index to remove
        """
        file_key = self.get_file_key(index)
        remove_file(file_key)

        stored_file_descriptions, stored_file_names, stored_file_sizes = self._get_metadata_from_block()

        stored_file_descriptions[index] = None
        self._set_file_descriptions(stored_file_descriptions)

        stored_file_names[index] = None
        self._set_file_names(stored_file_names)

        stored_file_sizes[index] = 0
        self._set_file_sizes(stored_file_sizes)

        if self.block.is_team_assignment():
            try:
                SharedFileUpload.by_key(file_key).delete()
            except SharedFileUpload.DoesNotExist:
                logger.warning('Could not find SharedFileUpload to delete: %s', file_key)

        self.invalidate_cached_shared_file_dicts()

    def _get_metadata_from_block(self):
        descriptions = self._get_file_descriptions()
        names = self._get_file_names(descriptions)
        sizes = self._get_file_sizes(descriptions)
        return descriptions, names, sizes

    def _file_uploads_from_list_fields(self, descriptions, names, sizes, include_deleted=False):
        """
        Given file upload data as list fields, return a list of FileUploads constructed from those fields
        """
        file_fields_by_key = {
            'name': names,
            'description': descriptions,
            'size': sizes,
        }

        if not descriptions:
            return self._descriptionless_uploads()

        file_uploads = []
        for index in range(len(descriptions)):
            file_upload_kwargs = {
                key: file_field[index] for key, file_field in file_fields_by_key.items()
            }

            file_upload_kwargs.update(self.student_item_dict)
            file_upload_kwargs['index'] = index

            file_upload = FileUpload(**file_upload_kwargs)
            if include_deleted or file_upload.exists:
                file_uploads.append(file_upload)

        return file_uploads

    def _descriptionless_uploads(self):
        """
        This is the old behavior, required for a corner case and should be eventually removed.
        https://github.com/edx/edx-ora2/pull/1275 closed a loophole that allowed files
        to be uploaded without descriptions. In that case, an ORA block's saved_file_descriptions would be
        an empty list, but a key corresponding to their student item information would exist (and thus,
        so would a valid download URL).
        If there are users in that state who have files uploaded
        with no descriptions but have not yet submitted, they will fall here.
        """
        file_uploads = []

        for index in range(self.block.MAX_FILES_COUNT):
            file_key = get_student_file_key(self.student_item_dict, index)

            download_url = ''
            try:
                download_url = get_download_url(file_key)
            except FileUploadError:
                pass

            if download_url:
                file_uploads.append(FileUpload(
                    name='', description='', size=0, index=index, descriptionless=True, **self.student_item_dict
                ))
            else:
                break

        return file_uploads

    def _get_file_descriptions(self):
        """ Returns a list of file descriptions associated with this manager's OA block. """
        return _safe_load_json_list(self.block.saved_files_descriptions)

    def _set_file_descriptions(self, file_description_list):
        """ Updates the file descriptions associated with this manager's OA block. """
        self.block.saved_files_descriptions = json.dumps(file_description_list)

    def _get_file_names(self, descriptions=None):
        """ Returns a list of file names associated with this manager's OA block. """
        descriptions = descriptions or self._get_file_descriptions()
        file_names = _safe_load_json_list(self.block.saved_files_names)
        if len(file_names) != len(descriptions):
            file_names = [None for _ in range(len(descriptions))]
            self._set_file_names(file_names)
        return file_names

    def _set_file_names(self, file_name_list):
        """ Updates the list of file names associated with this manager's OA block. """
        self.block.saved_files_names = json.dumps(file_name_list)

    def _get_file_sizes(self, descriptions=None):
        """ Returns a list of file sizes associated with this manager's OA block. """
        descriptions = descriptions or self._get_file_descriptions()
        file_sizes = _safe_load_json_list(self.block.saved_files_sizes)
        if len(file_sizes) != len(descriptions):
            file_sizes = [None for _ in range(len(descriptions))]
            self._set_file_sizes(file_sizes)
        return file_sizes

    def _set_file_sizes(self, file_size_list):
        self.block.saved_files_sizes = json.dumps(file_size_list)

    def _dicts_to_key_lists(self, dicts, required_keys):
        """
        Transposes a list of dictionaries with certain required keys
        to a tuple of lists, each containing the values of the required keys.
        """
        result = {
            key: [] for key in required_keys
        }

        for _dict in dicts:
            for key in required_keys:
                if key not in _dict:
                    raise FileUploadError(f'Missing required key {key} in {_dict}')
                result[key].append(_dict[key])

        return tuple(result[key] for key in required_keys)



# Generated by Django 2.2.10 on 2020-03-27 19:25

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('assessment', '0005_add_filename_to_sharedupload'),
    ]

    operations = [
        migrations.CreateModel(
            name='TeamStaffWorkflow',
            fields=[
                ('staffworkflow_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='assessment.StaffWorkflow')),
                ('team_submission_uuid', models.CharField(max_length=128, unique=True)),
            ],
            bases=('assessment.staffworkflow',),
        ),
    ]



"""
Provides the upload endpoint for the django storage backend.
"""


from django.contrib.auth.decorators import login_required
from django.shortcuts import HttpResponse
from django.views.decorators.http import require_http_methods

from .backends.django_storage import Backend


@login_required()
@require_http_methods(["PUT"])
def django_storage(request, key):
    """
    Upload files using django storage backend.
    """
    Backend().upload_file(key, request.body)
    return HttpResponse()



"""
Serializers for the Submission List endpoint
"""
from rest_framework import serializers
from openassessment.assessment.models.staff import StaffWorkflow


class MissingContextException(Exception):
    pass


# pylint: disable=abstract-method
class SubmissionListScoreSerializer(serializers.Serializer):
    pointsEarned = serializers.IntegerField(source='points_earned')
    pointsPossible = serializers.IntegerField(source='points_possible')


class SubmissionListSerializer(serializers.ModelSerializer):
    """
    Serialized info about an item returned from the submission list endpoint
    """
    class Meta:
        model = StaffWorkflow
        fields = [
            'submissionUuid',
            'dateSubmitted',
            'dateGraded',
            'gradingStatus',
            'lockStatus',
            'gradedBy',
            'username',
            'teamName',
            'score'
        ]
        read_only_fields = fields

    requires_context = True

    # Required context
    CONTEXT_ANON_ID_TO_USERNAME = 'anonymous_id_to_username'
    CONTEXT_SUB_TO_ASSESSMENT = 'submission_uuid_to_assessment'
    CONTEXT_SUB_TO_ANON_ID = 'submission_uuid_to_student_id'

    def _verify_required_context(self, context):
        """Verify that required individual or team context is present for serialization"""
        context_keys = set(context.keys())

        # Required context for individual submissions
        required_context = set([
            self.CONTEXT_ANON_ID_TO_USERNAME,
            self.CONTEXT_SUB_TO_ASSESSMENT,
            self.CONTEXT_SUB_TO_ANON_ID
        ])

        missing_context = required_context - context_keys
        if missing_context:
            raise ValueError(f"Missing required context {' ,'.join(missing_context)}")

    def __init__(self, *args, **kwargs):
        self._verify_required_context(kwargs.get('context', {}))
        super().__init__(*args, **kwargs)

    submissionUuid = serializers.CharField(source='submission_uuid')
    dateSubmitted = serializers.CharField(source='created_at')
    dateGraded = serializers.CharField(source='grading_completed_at')
    dateGraded = serializers.SerializerMethodField()
    gradingStatus = serializers.CharField(source='grading_status')
    lockStatus = serializers.CharField(source='lock_status')
    gradedBy = serializers.SerializerMethodField()
    username = serializers.SerializerMethodField()
    teamName = serializers.SerializerMethodField()
    score = serializers.SerializerMethodField()

    def _get_username_from_context(self, anonymous_user_id):
        try:
            return self.context[self.CONTEXT_ANON_ID_TO_USERNAME][anonymous_user_id]
        except KeyError as e:
            raise MissingContextException(f"Username not found for anonymous user id {anonymous_user_id}") from e

    def _get_anonymous_id_from_context(self, submission_uuid):
        try:
            return self.context[self.CONTEXT_SUB_TO_ANON_ID][submission_uuid]
        except KeyError as e:
            raise MissingContextException(
                f"No submitter anonymous user id found for submission uuid {submission_uuid}"
            ) from e

    def get_dateGraded(self, workflow):
        return str(workflow.grading_completed_at)

    def get_gradedBy(self, workflow):
        if workflow.scorer_id:
            return self._get_username_from_context(workflow.scorer_id)
        else:
            return None

    def get_username(self, workflow):
        return self._get_username_from_context(
            self._get_anonymous_id_from_context(workflow.identifying_uuid)
        )

    def get_teamName(self, workflow):  # pylint: disable=unused-argument
        # For individual submissions, this is intentionally empty
        return None

    def get_score(self, workflow):
        assessment = self.context[self.CONTEXT_SUB_TO_ASSESSMENT].get(workflow.identifying_uuid)
        if assessment:
            return SubmissionListScoreSerializer(assessment).data
        else:
            return {}


class TeamSubmissionListSerializer(SubmissionListSerializer):
    """
    Serialized info about an team item returned from the submission list endpoint
    """

    submissionUuid = serializers.CharField(source='team_submission_uuid')

    # Context keys
    CONTEXT_ANON_ID_TO_USERNAME = 'anonymous_id_to_username'
    CONTEXT_SUB_TO_ASSESSMENT = 'submission_uuid_to_assessment'
    CONTEXT_SUB_TO_TEAM_ID = 'team_submission_uuid_to_team_id'
    CONTEXT_TEAM_ID_TO_TEAM_NAME = 'team_id_to_team_name'

    REQUIRED_CONTEXT_KEYS = [
        CONTEXT_ANON_ID_TO_USERNAME,
        CONTEXT_SUB_TO_ASSESSMENT,
        CONTEXT_SUB_TO_TEAM_ID,
        CONTEXT_TEAM_ID_TO_TEAM_NAME,
    ]

    def _verify_required_context(self, context):
        """Verify that required team context is present for serialization"""
        context_keys = set(context.keys())

        # Required context for serialization
        required_context = set(self.REQUIRED_CONTEXT_KEYS)

        missing_context = required_context - context_keys
        if missing_context:
            raise ValueError(f"Missing required context {' ,'.join(missing_context)}")

    def _get_team_name_from_context(self, team_id):
        try:
            return self.context[self.CONTEXT_TEAM_ID_TO_TEAM_NAME][team_id]
        except KeyError as e:
            raise MissingContextException(f"Team name not found for team id {team_id}") from e

    def _get_team_id_from_context(self, team_submission_uuid):
        try:
            return self.context[self.CONTEXT_SUB_TO_TEAM_ID][team_submission_uuid]
        except KeyError as e:
            raise MissingContextException(
                f"No submitter anonymous user id found for team submission uuid {team_submission_uuid}"
            ) from e

    def get_username(self, workflow):  # pylint: disable=unused-argument
        # For team submissions, this is intentionally empty
        return None

    def get_teamName(self, workflow):
        return self._get_team_name_from_context(
            self._get_team_id_from_context(workflow.identifying_uuid)
        )



"""
    Open Assessment Template mixin helps out in generating different presets to display in studio.
"""

from django.utils.translation import gettext as _

from openassessment.xblock.defaults import (BLANK_ASSESSMENT_MODULES, PEER_ASSESSMENT_MODULES, SELF_ASSESSMENT_MODULES,
                                            SELF_TO_PEER_ASSESSMENT_MODULES, SELF_TO_STAFF_ASSESSMENT_MODULES,
                                            STAFF_ASSESSMENT_MODULES)


class OpenAssessmentTemplatesMixin:
    """
    This helps to get templates for different type of assessment that is
    offered.
    """

    VALID_ASSESSMENT_TYPES_DISPLAY_NAMES = {
        "peer-assessment": _("Peer Assessment Only"),
        "self-assessment": _("Self Assessment Only"),
        "staff-assessment": _("Staff Assessment Only"),
        "self-to-peer": _("Self Assessment to Peer Assessment"),
        "self-to-staff": _("Self Assessment to Staff Assessment")
    }

    VALID_ASSESSMENT_TYPES_ASSESSMENT_MODULE = {
        "self-assessment": SELF_ASSESSMENT_MODULES,
        "peer-assessment": PEER_ASSESSMENT_MODULES,
        "staff-assessment": STAFF_ASSESSMENT_MODULES,
        "self-to-peer": SELF_TO_PEER_ASSESSMENT_MODULES,
        "self-to-staff": SELF_TO_STAFF_ASSESSMENT_MODULES,
        "blank-assessment": BLANK_ASSESSMENT_MODULES
    }

    @classmethod
    def templates(cls):
        """
        Returns a list of dictionary field: value objects that describe possible templates.
        """
        templates = []
        for assesment_type, display_name in cls.VALID_ASSESSMENT_TYPES_DISPLAY_NAMES.items():
            template_id = assesment_type
            template = cls._create_template_dict(template_id, display_name)
            templates.append(template)
        return templates

    @classmethod
    def _create_template_dict(cls, template_id, display_name):
        """
        Creates a dictionary for serving various metadata for the template.

        Args:
            template_id(str): template id of what assessement template needs to be served.
            display_name(str): display name of template.

        Returns:
            A dictionary with proper keys to be consumed.
        """
        return {
            "template_id": template_id,
            "metadata": {
                "display_name": display_name,
            }
        }

    @classmethod
    def get_template(cls, template_id):
        """
        Helps to generate various option level template for ORA.

        Args:
            template_id(str): template id of what assessement template needs to be served.

        Returns:
            A dictionary of payload to be consumed by Studio.
        """
        rubric_assessments = cls._create_rubric_assessment_dict(template_id)
        return {
            "data": rubric_assessments
        }

    @classmethod
    def _create_rubric_assessment_dict(cls, template_id):
        """
        Creates a dictionary of parameters to be passed while creating ORA xblock.

        Args:
            template_id(str): template id of what assessement template needs to be served.

        Returns:
            A dictionary of payload to be consumed by Studio.
        """
        assessment_module = cls.VALID_ASSESSMENT_TYPES_ASSESSMENT_MODULE \
            .get(template_id)
        return {
            "rubric_assessments": assessment_module
        }



""" Exceptions for FileUploads. """


class FileUploadError(Exception):
    """An error related to uploading files

    This is the generic error raised when a file could not be uploaded.

    """


class FileUploadInternalError(FileUploadError):
    """An error internal to the File Upload API.

    This is an error raised when file upload failed due to internal problems in
    the File Upload API, beyond the intervention of the requester.

    """


class FileUploadRequestError(FileUploadError):
    """This error is raised when the request has invalid parameters for upload.

    This error will be raised if the file being uploaded is somehow invalid,
    based on type restrictions, size restrictions, upload limits, etc.

    """



""" A Mixin for Response submissions. """
import copy
import json
import logging
import os

from django.core.exceptions import ObjectDoesNotExist
from django.utils.functional import cached_property
from xblock.core import XBlock
from xblock.exceptions import NoSuchServiceError

from submissions.team_api import get_team_submission

from openassessment.fileupload import api as file_upload_api
from openassessment.fileupload.exceptions import FileUploadError
from openassessment.workflow.errors import AssessmentWorkflowError

from ..data import OraSubmissionAnswerFactory
from .data_conversion import (
    create_submission_dict,
    list_to_conversational_format,
    prepare_submission_for_serialization
)
from .resolve_dates import DISTANT_FUTURE
from .user_data import get_user_preferences
from .validation import validate_submission

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class NoTeamToCreateSubmissionForError(Exception):
    pass


class EmptySubmissionError(Exception):
    pass


class SubmissionMixin:
    """Submission Mixin introducing all Submission-related functionality.

    Submission Mixin contains all logic and handlers associated with rendering
    the submission section of the front end, as well as making all API calls to
    the middle tier for constructing new submissions, or fetching submissions.

    SubmissionMixin is a Mixin for the OpenAssessmentBlock. Functions in the
    SubmissionMixin call into the OpenAssessmentBlock functions and will not
    work outside the scope of OpenAssessmentBlock.

    """

    ALLOWED_IMAGE_MIME_TYPES = ['image/gif', 'image/jpeg', 'image/pjpeg', 'image/png']  # pragma: no cover
    ALLOWED_IMAGE_EXTENSIONS = ['gif', 'jpg', 'jpeg', 'jfif', 'pjpeg', 'pjp', 'png']  # pragma: no cover

    ALLOWED_FILE_MIME_TYPES = ['application/pdf'] + ALLOWED_IMAGE_MIME_TYPES  # pragma: no cover
    ALLOWED_FILE_EXTENSIONS = ['pdf'] + ALLOWED_IMAGE_EXTENSIONS  # pragma: no cover

    MAX_FILES_COUNT = 20  # pragma: no cover

    # taken from http://www.howtogeek.com/137270/50-file-extensions-that-are-potentially-dangerous-on-windows/
    # and http://pcsupport.about.com/od/tipstricks/a/execfileext.htm
    # left out .js and office extensions
    FILE_EXT_BLACK_LIST = [
        'exe', 'msi', 'app', 'dmg', 'com', 'pif', 'application', 'gadget',
        'msp', 'scr', 'hta', 'cpl', 'msc', 'jar', 'bat', 'cmd', 'vb', 'vbs',
        'jse', 'ws', 'wsf', 'wsc', 'wsh', 'scf', 'lnk', 'inf', 'reg', 'ps1',
        'ps1xml', 'ps2', 'ps2xml', 'psc1', 'psc2', 'msh', 'msh1', 'msh2', 'mshxml',
        'msh1xml', 'msh2xml', 'action', 'apk', 'app', 'bin', 'command', 'csh',
        'ins', 'inx', 'ipa', 'isu', 'job', 'mst', 'osx', 'out', 'paf', 'prg',
        'rgs', 'run', 'sct', 'shb', 'shs', 'u3p', 'vbscript', 'vbe', 'workflow',
        'htm', 'html',
    ]

    FILE_UPLOAD_PRESETS = {
        'image': {
            'mime_types': ALLOWED_IMAGE_MIME_TYPES,
            'extensions': ALLOWED_IMAGE_EXTENSIONS
        },
        'pdf-and-image': {
            'mime_types': ALLOWED_FILE_MIME_TYPES,
            'extensions': ALLOWED_FILE_EXTENSIONS,
        },
        'custom': {}
    }

    @cached_property
    def file_manager(self):
        return file_upload_api.FileUploadManager(self)

    @XBlock.json_handler
    def submit(self, data, suffix=''):  # pylint: disable=unused-argument
        """Place the submission text into Openassessment system

        Allows submission of new responses.  Performs basic workflow validation
        on any new submission to ensure it is acceptable to receive a new
        response at this time.

        Args:
            data (dict): Data may contain two attributes: submission and
                file_urls. submission is the response from the student which
                should be stored in the Open Assessment system. file_urls is the
                path to a related file for the submission. file_urls is optional.
            suffix (str): Not used in this handler.

        Returns:
            (tuple | [tuple]): Returns the status (boolean) of this request, the
                associated status tag (str), and status text (unicode).
                This becomes an array of similarly structured tuples in the event
                of a team submisison, one entry per student entry.

        """
        # Import is placed here to avoid model import at project startup.
        from submissions import api
        if 'submission' not in data:
            return (
                False,
                'EBADARGS',
                self._('"submission" required to submit answer.')
            )

        status = False
        student_sub_data = data['submission']
        success, msg = validate_submission(student_sub_data, self.prompts, self._, self.text_response)
        if not success:
            return (
                False,
                'EBADARGS',
                msg
            )

        student_item_dict = self.get_student_item_dict()

        # Short-circuit if no user is defined (as in Studio Preview mode)
        # Since students can't submit, they will never be able to progress in the workflow
        if self.in_studio_preview:
            return (
                False,
                'ENOPREVIEW',
                self._('To submit a response, view this component in Preview or Live mode.')
            )

        workflow = self.get_workflow_info()

        status_tag = 'ENOMULTI'  # It is an error to submit multiple times for the same item
        status_text = self._('Multiple submissions are not allowed.')

        if not workflow:
            try:

                # a submission for a team generates matching submissions for all members
                if self.is_team_assignment():
                    submission = self.create_team_submission(student_sub_data)
                else:
                    submission = self.create_submission(student_item_dict, student_sub_data)
                return self._create_submission_response(submission)

            except api.SubmissionRequestError as err:

                # Handle the case of an answer that's too long as a special case,
                # so we can display a more specific error message.
                # Although we limit the number of characters the user can
                # enter on the client side, the submissions API uses the JSON-serialized
                # submission to calculate length.  If each character submitted
                # by the user takes more than 1 byte to encode (for example, double-escaped
                # newline characters or non-ASCII unicode), then the user might
                # exceed the limits set by the submissions API.  In that case,
                # we display an error message indicating that the answer is too long.
                answer_too_long = any(
                    "maximum answer size exceeded" in answer_err.lower()
                    for answer_err in err.field_errors.get('answer', [])
                )
                if answer_too_long:
                    status_tag = 'EANSWERLENGTH'
                else:
                    msg = (
                        "The submissions API reported an invalid request error "
                        "when submitting a response for the user: {student_item}"
                    ).format(student_item=student_item_dict)
                    logger.exception(msg)
                    status_tag = 'EBADFORM'
                    status_text = msg
            except EmptySubmissionError:
                msg = (
                    "Attempted to submit submission for user {student_item}, "
                    "but submission contained no content."
                ).format(student_item=student_item_dict)
                logger.exception(msg)
                status_tag = 'EEMPTYSUB'
                status_text = self._(
                    'Submission cannot be empty. '
                    'Please refresh the page and try again.'
                )
            except (api.SubmissionError, AssessmentWorkflowError, NoTeamToCreateSubmissionForError):
                msg = (
                    "An unknown error occurred while submitting "
                    "a response for the user: {student_item}"
                ).format(student_item=student_item_dict)
                logger.exception(msg)
                status_tag = 'EUNKNOWN'
                status_text = self._('API returned unclassified exception.')

        # error cases fall through to here
        return status, status_tag, status_text

    def _create_submission_response(self, submission):
        """ Wrap submisison info for return to client

            Returns:
                (tuple): True (indicates success), student item, attempt number
        """
        status = True
        status_tag = submission.get('student_item')
        status_text = submission.get('attempt_number')

        return (status, status_tag, status_text)

    @XBlock.json_handler
    def save_submission(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Save the current student's response submission.
        If the student already has a response saved, this will overwrite it.

        Args:
            data (dict): Data should have a single key 'submission' that contains
                the text of the student's response. Optionally, the data could
                have a 'file_urls' key that is the path to an associated file for
                this submission.
            suffix (str): Not used.

        Returns:
            dict: Contains a bool 'success' and unicode string 'msg'.
        """
        if 'submission' in data:
            student_sub_data = data['submission']
            success, msg = validate_submission(student_sub_data, self.prompts, self._, self.text_response)
            if not success:
                return {'success': False, 'msg': msg}
            try:
                self.saved_response = json.dumps(
                    prepare_submission_for_serialization(student_sub_data)
                )
                self.has_saved = True

                # Emit analytics event...
                self.runtime.publish(
                    self,
                    "openassessmentblock.save_submission",
                    {"saved_response": self.saved_response}
                )
            except Exception:  # pylint: disable=broad-except
                return {'success': False, 'msg': self._("This response could not be saved.")}
            else:
                return {'success': True, 'msg': ''}
        else:
            return {'success': False, 'msg': self._("This response was not submitted.")}

    @XBlock.json_handler
    def save_files_descriptions(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Save the metadata for each uploaded file.

        Args:
            data (dict): Data should have a single key 'fileMetadata' that contains
                a list of dictionaries with the following keys: 'description','fileName', and 'fileSize'
            each element of the list maps to a single file
            suffix (str): Not used.

        Returns:
            dict: Contains a bool 'success' and unicode string 'msg'.
        """
        failure_response = {'success': False, 'msg': self._("Files descriptions were not submitted.")}

        if 'fileMetadata' not in data:
            return failure_response

        if not isinstance(data['fileMetadata'], list):
            return failure_response

        file_data = [
            {
                'description': item['description'],
                'name': item['fileName'],
                'size': item['fileSize'],
            } for item in data['fileMetadata']
        ]

        for new_upload in file_data:
            if not all([
                isinstance(new_upload['description'], str),
                isinstance(new_upload['name'], str),
                isinstance(new_upload['size'], int),
            ]):
                return failure_response

        try:
            self.file_manager.append_uploads(*file_data)
            # Emit analytics event...
            self.runtime.publish(
                self,
                "openassessmentblock.save_files_descriptions",
                {"saved_response": self.saved_files_descriptions}
            )
        except FileUploadError as exc:
            logger.exception(
                "FileUploadError: file description for data %s failed with error %s",
                data,
                exc,
                exc_info=True,
            )
            return {'success': False, 'msg': self._("Files metadata could not be saved.")}

        return {'success': True, 'msg': ''}

    def create_team_submission(self, student_sub_data):
        """ A student submitting for a team should generate matching submissions for every member of the team. """
        if not self.has_team():
            msg = "Student {} has no team for course {}".format(
                self.get_student_item_dict()['student_id'],
                self.course_id
            )
            logger.exception(msg)
            raise NoTeamToCreateSubmissionForError(msg)

        # Import is placed here to avoid model import at project startup.
        from submissions import team_api

        team_info = self.get_team_info()
        # Store the student's response text in a JSON-encodable dict
        # so that later we can add additional response fields.
        student_sub_dict = prepare_submission_for_serialization(student_sub_data)

        self._collect_files_for_submission(student_sub_dict)

        self.check_for_empty_submission_and_raise_error(student_sub_dict)

        submitter_anonymous_user_id = self.xmodule_runtime.anonymous_student_id
        user = self.get_real_user(submitter_anonymous_user_id)
        student_item_dict = self.get_student_item_dict(anonymous_user_id=submitter_anonymous_user_id)
        anonymous_student_ids = self.get_anonymous_user_ids_for_team()
        submission = team_api.create_submission_for_team(
            self.course_id,
            student_item_dict['item_id'],
            team_info['team_id'],
            user.id,
            anonymous_student_ids,
            student_sub_dict,
        )

        self.create_team_workflow(submission["team_submission_uuid"])
        # Emit analytics event...
        self.runtime.publish(
            self,
            "openassessmentblock.create_team_submission",
            {
                "submission_uuid": submission["team_submission_uuid"],
                "team_id": team_info["team_id"],
                "attempt_number": submission["attempt_number"],
                "created_at": submission["created_at"],
                "submitted_at": submission["submitted_at"],
                "answer": submission["answer"],
            }
        )
        return submission

    def create_submission(self, student_item_dict, student_sub_data):
        """ Creates submission for the submitted assessment response or a list for a team assessment. """
        # Import is placed here to avoid model import at project startup.
        from submissions import api

        # Store the student's response text in a JSON-encodable dict
        # so that later we can add additional response fields.
        student_sub_dict = prepare_submission_for_serialization(student_sub_data)

        self._collect_files_for_submission(student_sub_dict)

        self.check_for_empty_submission_and_raise_error(student_sub_dict)

        submission = api.create_submission(student_item_dict, student_sub_dict)
        self.create_workflow(submission["uuid"])
        self.submission_uuid = submission["uuid"]

        # Emit analytics event...
        self.runtime.publish(
            self,
            "openassessmentblock.create_submission",
            {
                "submission_uuid": submission["uuid"],
                "attempt_number": submission["attempt_number"],
                "created_at": submission["created_at"],
                "submitted_at": submission["submitted_at"],
                "answer": submission["answer"],
            }
        )

        return submission

    def check_for_empty_submission_and_raise_error(self, student_sub_dict):
        """
        Check if student_sub_dict has any submission content so that we don't
        create empty submissions.

        If there are no text responses and no file responses, raise an EmptySubmissionError
        """
        has_content = False

        # Does the student_sub_dict have any non-zero-length strings in 'parts'?
        has_content |= any(part.get('text', '') for part in student_sub_dict.get('parts', []))

        # Are there any file_keys in student_sub_dict?
        has_content |= len(student_sub_dict.get('file_keys', [])) > 0

        if not has_content:
            raise EmptySubmissionError

    def _collect_files_for_submission(self, student_sub_dict):
        """ Collect files from CSM for individual submisisons or SharedFileUpload for team submisisons. """

        if not self.file_upload_type:
            return None

        for field in ('file_keys', 'files_descriptions', 'files_names', 'files_sizes'):
            student_sub_dict[field] = []

        team_id = None if not self.has_team() else self.team.team_id
        uploads = self.file_manager.get_uploads(team_id=team_id)
        if self.is_team_assignment():
            uploads += self.file_manager.get_team_uploads(team_id=team_id)

        for upload in uploads:
            student_sub_dict['file_keys'].append(upload.key)
            student_sub_dict['files_descriptions'].append(upload.description)
            student_sub_dict['files_names'].append(upload.name)
            student_sub_dict['files_sizes'].append(upload.size)

        return student_sub_dict

    @XBlock.json_handler
    def get_student_username(self, data, suffix):  # pylint: disable=unused-argument
        """
        Gets the username of the current student for use in team lookup.
        """
        anonymous_id = self.xmodule_runtime.anonymous_student_id
        return {'username': self.get_username(anonymous_id)}

    @XBlock.json_handler
    def upload_url(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Request a URL to be used for uploading content related to this
        submission.

        Returns:
            A URL to be used to upload content associated with this submission.

        """
        if 'contentType' not in data or 'filename' not in data:
            return {'success': False, 'msg': self._("There was an error uploading your file.")}

        if not self.allow_multiple_files:

            # Here we check if there are existing file uploads by checking for
            # an existing download url for any of the upload slots.
            # Note that we can't use self.saved_files_descriptions because that
            # is populated before files are uploaded
            for i in range(self.MAX_FILES_COUNT):
                file_url = self._get_download_url(i)
                if file_url:
                    return {'success': False,
                            'msg': self._("Only a single file upload is allowed for this assessment.")}

        file_num = int(data.get('filenum', 0))

        _, file_ext = os.path.splitext(data['filename'])
        file_ext = file_ext.strip('.') if file_ext else None
        content_type = data['contentType']

        # Validate that there are no data issues and file type is allowed
        if not self.is_supported_upload_type(file_ext, content_type):
            return {'success': False, 'msg': self._(
                "File upload failed: unsupported file type."
                "Only the supported file types can be uploaded."
                "If you have questions, please reach out to the course team."
            )}

        # Attempt to upload
        file_num = int(data.get('filenum', 0))
        try:
            key = self._get_student_item_key(file_num)
            url = file_upload_api.get_upload_url(key, content_type)
            return {'success': True, 'url': url}
        except FileUploadError:
            logger.exception("FileUploadError:Error retrieving upload URL for the data: %s.", data)
            return {'success': False, 'msg': self._("Error retrieving upload URL.")}

    def is_supported_upload_type(self, file_ext, content_type):
        """
        Determine if the uploaded file type/extension is allowed for the configured file upload configuration

        Returns:
            True/False if file type is supported/unsupported
        """
        if self.file_upload_type == 'image' and content_type not in self.ALLOWED_IMAGE_MIME_TYPES:
            return False

        elif self.file_upload_type == 'pdf-and-image' and content_type not in self.ALLOWED_FILE_MIME_TYPES:
            return False

        elif self.file_upload_type == 'custom' and file_ext.lower() not in self.white_listed_file_types:
            return False

        elif file_ext in self.FILE_EXT_BLACK_LIST:
            return False

        return True

    @XBlock.json_handler
    def download_url(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Request a download URL.

        Returns:
            A URL to be used for downloading content related to the submission.

        """
        file_num = int(data.get('filenum', 0))
        return {'success': True, 'url': self._get_download_url(file_num)}

    @XBlock.json_handler
    def remove_uploaded_file(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Removes uploaded user file.
        """
        filenum = data.get('filenum', -1)
        try:
            filenum = int(filenum)
        except ValueError:
            filenum = -1
        student_item_key = self._get_student_item_key(num=filenum)
        if self._can_delete_file(filenum):
            try:
                self.file_manager.delete_upload(filenum)
                # Emit analytics event...
                self.runtime.publish(
                    self,
                    "openassessmentblock.remove_uploaded_file",
                    {"student_item_key": student_item_key}
                )
                logger.debug("Deleted file %s", student_item_key)
                return {'success': True}
            except FileUploadError as exc:
                logger.exception(
                    "FileUploadError: Error when deleting file %s : %s",
                    student_item_key,
                    exc,
                    exc_info=True
                )

        return {'success': False}

    def _can_delete_file(self, filenum):
        """
        Helper function, wraps `file_upload_api.can_delete_file()`.
        """
        team_id = self.get_team_info().get('team_id')
        key = self._get_student_item_key(filenum)
        current_user_id = self.get_student_item_dict()['student_id']
        return file_upload_api.can_delete_file(current_user_id, self.teams_enabled, key, team_id)

    def _get_download_url(self, file_num=0):
        """
        Internal function for retrieving the download url.

        """
        return self._get_url_by_file_key(self._get_student_item_key(file_num))

    def _get_student_item_key(self, num=0):
        """
        Simple utility method to generate a common file upload key based on
        the student item.

        Returns:
            A string representation of the key.

        """
        return file_upload_api.get_student_file_key(self.get_student_item_dict(), index=num)

    @classmethod
    def _get_url_by_file_key(cls, key):
        """
        Return download url for some particular file key.

        """
        url = ''
        try:
            if key:
                url = file_upload_api.get_download_url(key)
        except FileUploadError as exc:
            logger.exception(
                "FileUploadError: Download url for file key %s failed with error %s",
                key,
                exc,
                exc_info=True
            )

        return url

    @classmethod
    def get_download_urls_from_submission(cls, submission):
        """
        Returns a download URLs for retrieving content within a submission.

        Args:
            submission (dict): Dictionary containing an answer and a file_keys.
                The file_keys is used to try and retrieve a download urls
                with related content

        Returns:
            List of FileDescriptor dicts for each file associated with the submission

        """
        urls = []
        raw_answer = submission.get('answer')
        answer = OraSubmissionAnswerFactory.parse_submission_raw_answer(raw_answer)
        for file_upload in answer.get_file_uploads(missing_blank=True):
            file_download_url = cls._get_url_by_file_key(file_upload.key)
            if file_download_url:
                urls.append(
                    file_upload_api.FileDescriptor(
                        download_url=file_download_url,
                        description=file_upload.description,
                        name=file_upload.name,
                        size=file_upload.size,
                        show_delete_button=False
                    )._asdict()
                )
        return urls

    def get_files_info_from_user_state(self, username):
        """
        Returns the files information from the user state for a given username.

        If the files information is present in the user state, return a list of following tuple:
        (file_download_url, file_description, file_name)

        Arguments:
            username(str): user's name whose state is being check for files information.
        Returns:
            List of FileDescriptor dicts, if present, else empty list.
        """

        files_info = []
        user_state = self.get_user_state(username)
        item_dict = self.get_student_item_dict_from_username_or_email(username)
        if 'saved_files_descriptions' in user_state:
            # pylint: disable=protected-access
            files_descriptions = file_upload_api._safe_load_json_list(
                user_state.get('saved_files_descriptions'),
                log_error=True
            )
            files_names = file_upload_api._safe_load_json_list(
                user_state.get('saved_files_names', '[]'),
                log_error=True
            )
            for index, description in enumerate(files_descriptions):
                file_key = file_upload_api.get_student_file_key(item_dict, index)
                download_url = self._get_url_by_file_key(file_key)
                if download_url:
                    file_name = files_names[index] if index < len(files_names) else ''
                    files_info.append(
                        file_upload_api.FileDescriptor(
                            download_url=download_url,
                            description=description,
                            name=file_name,
                            size=None,
                            show_delete_button=False
                        )._asdict()
                    )
                else:
                    # If file has been removed, the URL doesn't exist
                    logger.info(
                        "URLWorkaround: no URL for description %s & key %s for user:%s",
                        description,
                        username,
                        file_key
                    )
                    continue
        return files_info

    def get_all_upload_urls_for_user(self, username_or_email):
        """
        For a particular ORA block, get the download URLs for all the files uploaded and still present.

        Used for an extreme edge case, where the stored files indices are out of sync with
        the uploaded files, this is a last resort to get the download URLs of all the files
        that have been uploaded by a learner in an ORA block(and haven't been deleted from the storage).
        Starting from 0 index to maximum file upload count possible, this checks if a file exists against
        every index. If present, add the info, else repeat it for the next indices.

        Arguments:
            username_or_email(str): username or email of the learner whose files' information is to be obtained.
        Returns:
            List of FileDescriptor dicts
        """
        file_uploads = []
        student_item_dict = self.get_student_item_dict_from_username_or_email(username_or_email)
        for index in range(self.MAX_FILES_COUNT):
            file_key = file_upload_api.get_student_file_key(student_item_dict, index)
            download_url = ''
            try:
                download_url = file_upload_api.get_download_url(file_key)
            except FileUploadError:
                pass

            if download_url:
                logger.info(
                    "Download URL exists for key %s in block %s for user %s",
                    file_key,
                    username_or_email,
                    str(self.location)
                )
                file_uploads.append(
                    file_upload_api.FileDescriptor(
                        download_url=download_url,
                        description='',
                        name='',
                        size=None,
                        show_delete_button=False
                    )._asdict()
                )
            else:
                continue

        return file_uploads

    @staticmethod
    def get_user_submission(submission_uuid):
        """Return the most recent submission by user in workflow

        Return the most recent submission.  If no submission is available,
        return None. All submissions are preserved, but only the most recent
        will be returned in this function, since the active workflow will only
        be concerned with the most recent submission.

        Args:
            submission_uuid (str): The uuid for the submission to retrieve.

        Returns:
            (dict): A dictionary representation of a submission to render to
                the front end.

        """
        # Import is placed here to avoid model import at project startup.
        from submissions import api
        try:
            return api.get_submission(submission_uuid)
        except api.SubmissionRequestError:
            # This error is actually ok.
            return None

    @property
    def save_status(self):
        """
        Return a string indicating whether the response has been saved.

        Returns:
            unicode
        """
        return self._('This response has been saved but not submitted.') if self.has_saved else self._(
            'This response has not been saved.')

    @XBlock.handler
    def render_submission(self, data, suffix=''):  # pylint: disable=unused-argument
        """Renders the Submission HTML section of the XBlock

        Generates the submission HTML for the first section of an Open
        Assessment XBlock. See OpenAssessmentBlock.render_assessment() for
        more information on rendering XBlock sections.

        Needs to support the following scenarios:
        Unanswered and Open
        Unanswered and Closed
        Saved
        Saved and Closed
        Submitted
        Submitted and Closed
        Submitted, waiting assessment
        Submitted and graded

        """
        path, context = self.submission_path_and_context()
        return self.render_assessment(path, context_dict=context)

    def get_team_submission_context(self, context):
        """
        Populate the passed context object with team info, including a set of students on
        the team with submissions to the current item from another team, under the key
        `team_members_with_external_submissions`.

        Args:
            context (dict): render context to add team submission context into
        Returns
            (dict): context arg with additional team-related fields
        """

        from submissions import team_api
        try:
            team_info = self.get_team_info()
            if team_info:
                context.update(team_info)
                if self.is_course_staff:
                    return
                student_item_dict = self.get_student_item_dict()
                external_submissions = team_api.get_teammates_with_submissions_from_other_teams(
                    self.course_id,
                    student_item_dict["item_id"],
                    team_info["team_id"],
                    self.get_anonymous_user_ids_for_team()
                )

                context["team_members_with_external_submissions"] = list_to_conversational_format([
                    self.get_username(submission['student_id']) for submission in external_submissions
                ])
        except ObjectDoesNotExist:
            logger.error(
                '%s: User associated with anonymous_user_id %s can not be found.',
                str(self.location),
                self.get_student_item_dict()['student_id'],
            )
        except NoSuchServiceError:
            logger.error('%s: Teams service is unavailable', str(self.location))

    def get_allowed_file_types_or_preset(self):
        """
        If allowed files are not explicitly set for file uploads, use preset extensions
        """
        if self.white_listed_file_types:
            return self.white_listed_file_types
        elif self.file_upload_type == 'image':
            return self.ALLOWED_IMAGE_EXTENSIONS
        elif self.file_upload_type == 'pdf-and-image':
            return self.ALLOWED_FILE_EXTENSIONS
        return None

    def submission_path_and_context(self):
        """
        Determine the template path and context to use when
        rendering the response (submission) step.

        Returns:
            tuple of `(path, context)`, where `path` (str) is the path to the template,
            and `context` (dict) is the template context.

        """
        workflow = self.get_team_workflow_info() if self.teams_enabled else self.get_workflow_info()
        problem_closed, reason, start_date, due_date = self.is_closed('submission')
        user_preferences = get_user_preferences(self.runtime.service(self, 'user'))

        path = 'openassessmentblock/response/oa_response.html'
        context = {
            'enable_delete_files': False,
            'file_upload_response': self.file_upload_response,
            'has_real_user': self.has_real_user,
            'prompts_type': self.prompts_type,
            'show_rubric_during_response': self.show_rubric_during_response,
            'text_response': self.text_response,
            'text_response_editor': self.text_response_editor,
            'user_language': user_preferences['user_language'],
            'user_timezone': user_preferences['user_timezone'],
            'xblock_id': self.get_xblock_id(),
        }

        if self.show_rubric_during_response:
            context['rubric_criteria'] = copy.deepcopy(self.rubric_criteria_with_labels)

        # Due dates can default to the distant future, in which case
        # there's effectively no due date.
        # If we don't add the date to the context, the template won't display it.
        if due_date < DISTANT_FUTURE:
            context["submission_due"] = due_date

        # For team assignments, if a user submitted with a past team, that gets precidence.
        # So we first see if they have a submission and load context from that.
        # Otherwise, we fall back to the current team.
        team_id_for_current_submission = None
        if self.is_team_assignment():
            if not workflow:
                team_id_for_current_submission = self.get_team_info().get('team_id', None)
            else:
                team_submission = get_team_submission(workflow['team_submission_uuid'])
                team_id_for_current_submission = team_submission['team_id']

            # If it's a team assignment, the user hasn't submitted and is not on a team, the assignment is unavailable.
            if team_id_for_current_submission is None:
                path = 'openassessmentblock/response/oa_response_unavailable.html'
                return path, context

        context['file_upload_type'] = self.file_upload_type
        context['allow_multiple_files'] = self.allow_multiple_files
        context['allow_latex'] = self.allow_latex

        file_urls = None

        if self.file_upload_type:
            context['file_urls'] = self.file_manager.file_descriptors(
                team_id=team_id_for_current_submission, include_deleted=True
            )
            context['team_file_urls'] = self.file_manager.team_file_descriptors(
                team_id=team_id_for_current_submission
            )
            context['white_listed_file_types'] = ['.' + ext for ext in self.get_allowed_file_types_or_preset()]

        if not workflow and problem_closed:
            if reason == 'due':
                path = 'openassessmentblock/response/oa_response_closed.html'
            elif reason == 'start':
                context['submission_start'] = start_date
                path = 'openassessmentblock/response/oa_response_unavailable.html'
        elif not workflow:
            # For backwards compatibility. Initially, problems had only one prompt
            # and a string answer. We convert it to the appropriate dict.
            no_workflow_path = "openassessmentblock/response/oa_response.html"

            try:
                json.loads(self.saved_response)
                saved_response = {
                    'answer': json.loads(self.saved_response),
                }
            except ValueError:
                saved_response = {
                    'answer': {
                        'text': self.saved_response,
                    },
                }

            context['saved_response'] = create_submission_dict(saved_response, self.prompts)
            context['save_status'] = self.save_status
            context['enable_delete_files'] = True

            submit_enabled = True
            if self.text_response == 'required' and not self.saved_response:
                submit_enabled = False
            if self.file_upload_response == 'required' and not file_urls:
                submit_enabled = False
            if self.text_response == 'optional' and self.file_upload_response == 'optional' \
                    and not self.saved_response and not file_urls:
                submit_enabled = False
            context['submit_enabled'] = submit_enabled

            if self.teams_enabled:
                self.get_team_submission_context(context)
                if self.does_team_have_submission(context['team_id']):
                    no_workflow_path = 'openassessmentblock/response/oa_response_team_already_submitted.html'

            path = no_workflow_path
        elif workflow["status"] == "cancelled":
            if self.teams_enabled:
                context["workflow_cancellation"] = self.get_team_workflow_cancellation_info(
                    workflow["team_submission_uuid"])
            else:
                context["workflow_cancellation"] = self.get_workflow_cancellation_info(
                    self.submission_uuid)
            context["student_submission"] = self.get_user_submission(
                workflow["submission_uuid"]
            )
            path = 'openassessmentblock/response/oa_response_cancelled.html'
        elif workflow["status"] == "done":
            student_submission = self.get_user_submission(
                workflow["submission_uuid"]
            )
            context["student_submission"] = create_submission_dict(student_submission, self.prompts)
            path = 'openassessmentblock/response/oa_response_graded.html'
        else:
            student_submission = self.get_user_submission(
                workflow["submission_uuid"]
            )
            peer_in_workflow = "peer" in workflow["status_details"]
            self_in_workflow = "self" in workflow["status_details"]
            context["peer_incomplete"] = peer_in_workflow and not workflow["status_details"]["peer"]["complete"]
            context["self_incomplete"] = self_in_workflow and not workflow["status_details"]["self"]["complete"]
            context["student_submission"] = create_submission_dict(student_submission, self.prompts)
            path = 'openassessmentblock/response/oa_response_submitted.html'

        return path, context



""" Assessment Urls. """

from django.urls import re_path

from openassessment.assessment import views

urlpatterns = [
    re_path(
        r'^(?P<student_id>[^/]+)/(?P<course_id>[^/]+)/(?P<item_id>[^/]+)$',
        views.get_evaluations_for_student_item
    ),
]



'''
Add in /edx/app/edxapp/edx-platform/lms/envs/aws.py:
ORA2_SWIFT_URL = AUTH_TOKENS["ORA2_SWIFT_URL"]
ORA2_SWIFT_KEY = AUTH_TOKENS["ORA2_SWIFT_KEY"]

Add in /edx/app/edxapp/lms.auth.json
"ORA2_SWIFT_URL": "https://EXAMPLE",
"ORA2_SWIFT_KEY": "EXAMPLE",

ORA2_SWIFT_KEY should correspond to Meta Temp-Url-Key configure in swift. Run
'swift stat -v' to get it.
'''

import logging

import urllib
import requests

from django.conf import settings

import swiftclient

from ..exceptions import FileUploadInternalError
from .base import BaseBackend

logger = logging.getLogger("openassessment.fileupload.api")  # pylint: disable=invalid-name

# prefix paths with current version, in case we need to roll it at some point
SWIFT_BACKEND_VERSION = 1


class Backend(BaseBackend):
    """
    Upload openassessment student files to swift
    """

    def get_upload_url(self, key, content_type):
        bucket_name, key_name = self._retrieve_parameters(key)
        key, url = get_settings()
        try:
            temp_url = swiftclient.utils.generate_temp_url(
                path=f'/v{SWIFT_BACKEND_VERSION}{url.path}/{bucket_name}/{key_name}',
                key=key,
                method='PUT',
                seconds=self.UPLOAD_URL_TIMEOUT
            )
            return f'{url.scheme}://{url.netloc}{temp_url}'
        except Exception as ex:
            logger.exception(
                "An internal exception occurred while generating an upload URL."
            )
            raise FileUploadInternalError(ex) from ex

    def get_download_url(self, key):
        bucket_name, key_name = self._retrieve_parameters(key)
        key, url = get_settings()
        try:
            temp_url = swiftclient.utils.generate_temp_url(
                path=f'/v{SWIFT_BACKEND_VERSION}{url.path}/{bucket_name}/{key_name}',
                key=key,
                method='GET',
                seconds=self.DOWNLOAD_URL_TIMEOUT
            )
            download_url = f'{url.scheme}://{url.netloc}{temp_url}'
            response = requests.get(download_url)
            return download_url if response.status_code == 200 else ""
        except Exception as ex:
            logger.exception(
                "An internal exception occurred while generating a download URL."
            )
            raise FileUploadInternalError(ex) from ex

    def remove_file(self, key):
        bucket_name, key_name = self._retrieve_parameters(key)
        key, url = get_settings()
        try:
            temp_url = swiftclient.utils.generate_temp_url(
                path=f'{url.path}/{bucket_name}/{key_name}',
                key=key,
                method='DELETE',
                seconds=self.DOWNLOAD_URL_TIMEOUT)
            remove_url = f'{url.scheme}://{url.netloc}{temp_url}'
            response = requests.delete(remove_url)
            return response.status_code == 204
        except Exception as ex:
            logger.exception(
                "An internal exception occurred while removing object on swift storage."
            )
            raise FileUploadInternalError(ex) from ex


def get_settings():
    """
    Returns the swift key and a parsed url.
    Both are generated from django settings.
    """

    url = getattr(settings, 'ORA2_SWIFT_URL', None)
    key = getattr(settings, 'ORA2_SWIFT_KEY', None)
    url = urllib.parse.urlparse(url)
    return key, url



"""
Errors around Submission Locking
"""


class SubmissionLockContestedError(Exception):
    """
    Error indicating trying to modify a lock that the user does not have access to modify.
    """
    error_code = 'ERR_LOCK_CONTESTED'

    def get_error_code(self):
        return self.error_code



"""
Management command to create submissions, assessments, and locks to make testing easier.
"""
import logging
import json
from os.path import exists, join

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError
from django.core.management import call_command

from opaque_keys.edx.keys import CourseKey

import loremipsum
from submissions import api as sub_api
from openassessment.assessment.api import staff as staff_api
from openassessment.workflow import api as workflow_api
from openassessment.xblock.data_conversion import create_rubric_dict
from openassessment.staffgrader.models import SubmissionGradingLock

log = logging.getLogger('create_oa_submissions_from_file')

SUPERUSER_USERNAME = 'edx'

EPILOG = """
{
    "displayName": <Display Name for the target ORA>,
    "submissions": [
        {
            "username": <Username of submitter>,
            "lockOwner": <Username of lock owner, or null for no lock>,
            "gradeData":{
                "gradedBy": <Username of grader>,
                "overallFeedback": <Overall feedback>,
                "criteria": [ # Note: There must be a criterion for all criteria in the ORA rubric
                    {
                        "label": <Criterion label>,
                        "selectedOption": <label of selected option for this criterion>,
                        "feedback": <feedback for criterion>,
                    },
                    ...
                ]
            }
        },
        ...
    ]
}

The input file can be a single JSON object describing a single ORA or it can be a list of multiple objects
describing multiple ORAs in one course. If you have multiple ORAs specified, they must have unique display
names or the command won't know which is which. Also, any ORA display name must be unique *withon the course*.
Display name is the only way we currently identify blocks in this command so there bust be uniqueness.

Unfortunately due to how argparse works this will have no formatting on the command line, check this management
command's source for a formatted version.
"""


def generate_lorem_sentences(num_sentences=4):
    """
    Generate some number of sentences of lorem ipsum
    """
    sentences = [generate_lorem_sentence() for _ in range(num_sentences)]
    return " ".join(sentences)


def generate_lorem_sentence():
    """
    Generate a sentence of lorem ipsum. The loremipsum library is broken and returns things like
    "B'lorem' b'ipsum' b'next' b'nextword'."
    This just trims off the byte string markers that we're getting.
    """
    result = []
    lorem_sentence = loremipsum.get_sentence().split()
    for word in lorem_sentence:
        if word[-1] != "'":
            end_trim_index = len(word) - 2
        else:
            end_trim_index = len(word) - 1
        result.append(word[2:end_trim_index])
    return " ".join(result) + '.'


class Command(BaseCommand):
    """
    Management command to create submissions, assessments, and locks to make testing easier.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.username_to_anonymous_user_id = None
        self.display_name_to_block = None

    def add_arguments(self, parser):
        parser.add_argument(
            'course_id',
            type=CourseKey.from_string,
            help='The id of the course for which to create submissions'
        )
        parser.add_argument(
            'submissions_config_file_path',
            help=(
                'The path to the file describing the format and structure of the submissions and assessments to '
                'generate. The path should be relative from the the ora2 src directory. Input files should be placed '
                'somewhere in WORKSPACE_ROOT/src/edx-ora2/ on local disk, which corresponds to /edx/src/edx-ora2 in '
                'the edx-platform docker container. \n For file format, call this management command with --help to '
                'view a description of the file format.'
            )
        )
        modes_group = parser.add_mutually_exclusive_group(required=True)
        modes_group.add_argument(
            '--init',
            action='store_true',
            help='Create and enroll users, create submissions, assessments, and locks.'
        )
        modes_group.add_argument(
            '--reset',
            action='store_true',
            help='Clear all ORA submission state from the given course.'
        )
        modes_group.add_argument(
            '--submit',
            action='store_true',
            help='Create submissions, assessments, and locks.'
        )
        parser.epilog = EPILOG

    def handle(self, *args, **options):
        """
        Run the command. Do additional arg checking, parse and load the input file,
        do up-front loading, and then do either init, reset, or submit.
        """
        course_id = options['course_id']
        submissions_config = self.read_config_file(options['submissions_config_file_path'])
        if options['init']:
            self.init_ora_test_data(course_id, submissions_config)

        # We have to do the init first because otherwise the users won't exist, and we won't
        # be able to look up their anonymous ids.
        self.load_anonymous_ids_and_block_locations(course_id, submissions_config)

        if options['reset']:
            self.reset_ora_test_data(course_id, submissions_config)
        elif options['submit'] or options['init']:
            self.submit_ora_test_data(course_id, submissions_config)

    def read_config_file(self, file_path):
        """
        Check that the input file exits, and attempt to open and json parse the file.
        Returns the json parsed input file.
        """
        file_path = join('/' 'edx', 'src', 'edx-ora2', file_path)
        if not exists(file_path):
            raise CommandError(f'File {file_path} not found.')

        with open(file_path, 'r') as f:
            try:
                submissions_config = json.load(f)
                if not isinstance(submissions_config, list):
                    submissions_config = [submissions_config]
                return submissions_config
            except json.JSONDecodeError as e:
                raise CommandError(f'Unable to parse file {file_path}: {e}') from e

    def load_anonymous_ids_and_block_locations(self, course_id, submissions_config):
        """
        Do all upfront database and modulestore lookups.
        Look up anonymous user ids and grab ORA blocks from the modulestore.
        """
        learners, course_staff = self.get_usernames(submissions_config)
        self.load_anonymous_user_ids(course_id, learners | course_staff)

        ora_display_names = self.get_display_names(submissions_config)
        self.load_ora_blocks(course_id, ora_display_names)

    def get_usernames(self, submissions_config):
        """
        Go through the submissions config and gather all usernames into two sets:
        -usernames of people who submitted (learners)
        -usernames mentioned in the file as graders or lock owners (course staff)
        """
        learners = set()
        course_staff = set()
        for ora_config in submissions_config:
            for submission in ora_config['submissions']:
                learners.add(submission['username'])
                if submission['gradeData']:
                    course_staff.add(submission['gradeData']['gradedBy'])
                if submission['lockOwner'] is not None:
                    course_staff.add(submission['lockOwner'])

        return learners, course_staff

    def load_anonymous_user_ids(self, course_id, usernames):
        """
        Look up all Users with the given usernames, look up their anonymous user ids for the specified course, and
        store a mapping from username to anonaymous id in self.username_to_anonymous_user_id
        """
        try:
            from common.djangoapps.student.models import anonymous_id_for_user  # pylint: disable=import-error
        except ModuleNotFoundError as e:
            raise CommandError((
                "Cannot import common.djangoapps.student.models.anonymous_id_for_user. "
                "This management command must be run from the LMS shell."
            )) from e
        # Also include a superuser because when we reset we need to include a "reset by" anonymous id
        usernames.add(SUPERUSER_USERNAME)

        anonymous_user_ids = {}
        users = get_user_model().objects.filter(username__in=usernames).all()
        for user in users:
            anonymous_user_ids[user.username] = anonymous_id_for_user(user=user, course_id=course_id)

        missing_usernames = set(usernames) - set(anonymous_user_ids.keys())
        if missing_usernames:
            raise CommandError("Unable to load anonymous id for user(s) " + ' '.join(missing_usernames))

        self.username_to_anonymous_user_id = anonymous_user_ids

    def get_display_names(self, submissions_config):
        """
        Returns a list of all ORA display names mentioned in the submissions config
        """
        display_names = []
        for ora_config in submissions_config:
            display_name = ora_config['displayName']
            if display_name in display_names:
                raise CommandError('Duplicate ORA display name found in configuration file:' + display_name)
            display_names.append(display_name)
        return display_names

    def load_ora_blocks(self, course_id, display_names):
        """
        Look up openassessment blocks for the course from the modulestore, and save the ones that match the given
        display names in a dict mapping from display name to block in self.display_name_to_block
        """
        try:
            from xmodule.modulestore.django import modulestore  # pylint: disable=import-error
        except ModuleNotFoundError as e:
            raise CommandError((
                "Cannot import xmodule.modulestore.django.modulestore. "
                "This management command must be run from the LMS shell."
            )) from e
        openassessment_blocks = modulestore().get_items(
            course_id, qualifiers={'category': 'openassessment'}
        )
        display_name_to_block = {}
        for block in openassessment_blocks:
            if block.parent is not None and block.display_name in display_names:
                if block.display_name in display_name_to_block:
                    raise CommandError((
                        f"The ORA '{block.display_name}' is specified in the input file. "
                        "The course contains more than one ORA with that display name. "
                        f"First two found:  [{display_name_to_block[block.display_name]}, {block.location}]"
                    ))
                display_name_to_block[block.display_name] = block

        missing_display_names = set(display_names) - set(display_name_to_block.keys())
        if missing_display_names:
            raise CommandError(
                f"The following Display Name(s) were not found in {str(course_id)} {' '.join(missing_display_names)}"
            )
        self.display_name_to_block = display_name_to_block

    def init_ora_test_data(self, course_id, submissions_config):
        """
        Run the initialization. Create all users and enroll in the course.
        """
        learners, course_staff = self.get_usernames(submissions_config)

        log.info('Creating and enrolling %d learners', len(learners))
        call_command(
            'create_test_users',
            *learners,
            course=course_id,
            ignore_user_already_exists=True
        )

        log.info('Creating and enrolling %d course staff', len(course_staff))
        call_command(
            'create_test_users',
            *course_staff,
            course=course_id,
            course_staff=True,
            ignore_user_already_exists=True
        )

    def submit_ora_test_data(self, course_id, submissions_config):
        """
        Run the submit action. For each specified submission, create the submission, create an assessment if specified,
        and create a lock if specified.
        """
        for ora_config in submissions_config:
            log.info('Creating test submissions for course %s', course_id)
            for submission_config in ora_config['submissions']:
                log.info("Creating submission for user %s", submission_config['username'])
                student_item = self.student_item(
                    submission_config['username'],
                    course_id,
                    ora_config['displayName']
                )
                # Submissions consist of username, a line break, and then some lorem
                text_response = submission_config['username'] + '\n' + generate_lorem_sentences()
                submission = sub_api.create_submission(student_item, {'parts': [{'text': text_response}]})
                workflow_api.create_workflow(submission['uuid'], ['staff'])
                workflow_api.update_from_assessments(submission['uuid'], None)
                log.info("Created submission %s for user %s", submission['uuid'], submission_config['username'])

                if submission_config['lockOwner']:
                    log.info(
                        "Creating lock on submission %s owned by %s",
                        submission['uuid'],
                        submission_config['lockOwner']
                    )
                    SubmissionGradingLock.claim_submission_lock(
                        submission['uuid'],
                        self.username_to_anonymous_user_id[submission_config['lockOwner']]
                    )

                if submission_config['gradeData']:
                    grade_data = submission_config['gradeData']
                    log.info(
                        "Creating assessment from user %s for submission %s",
                        grade_data['gradedBy'],
                        submission['uuid']
                    )
                    block = self.display_name_to_block[ora_config['displayName']]
                    rubric_dict = create_rubric_dict(block.prompts, block.rubric_criteria_with_labels)
                    options_selected, criterion_feedback = self.api_format_criteria(grade_data['criteria'], rubric_dict)
                    staff_api.create_assessment(
                        submission['uuid'],
                        self.username_to_anonymous_user_id[grade_data['gradedBy']],
                        options_selected,
                        criterion_feedback,
                        grade_data['overallFeedback'],
                        rubric_dict,
                    )
                    workflow_api.update_from_assessments(submission['uuid'], None)

    def student_item(self, username, course_id, ora_display_name):
        """Helper for creating student item dicts"""
        return {
            'student_id': self.username_to_anonymous_user_id[username],
            'course_id': str(course_id),
            'item_id': str(self.display_name_to_block[ora_display_name].location),
            'item_type': 'openassessment'
        }

    def lookup_criterion_and_option_name(self, criterion_label, option_label, rubric_dict):
        """
        The label that users see in Studio for criteria and options are the LABELs not the NAMEs.
        The API expects names, and we have labels, so we need to look at the block's rubric definition to convert.
        """
        for criterion in rubric_dict['criteria']:
            if criterion['label'] == criterion_label:
                criterion_name = criterion['name']
                for option in criterion['options']:
                    if option['label'] == option_label:
                        return criterion_name, option['name']
        raise ValueError(f"Can't find criterion and option names for labels {criterion_label}, {option_label}")

    def api_format_criteria(self, criteria, rubric_dict):
        """
        Our input file is specifying assessments as a list of objects that link one criterion with the selected
        option name and any feedback for that criterion.
        The API wants two dicts, one from criteria name to selected option name, and one from criteria name to
        feedback for that criterion.
        """
        options_selected = {}
        criterion_feedback = {}
        for criterion in criteria:
            criterion_label = criterion['label']
            option_label = criterion['selectedOption']
            criterion_name, option_name = self.lookup_criterion_and_option_name(
                criterion_label, option_label, rubric_dict
            )
            options_selected[criterion_name] = option_name
            feedback = criterion.get('feedback')
            if feedback is not None:
                criterion_feedback[criterion_name] = feedback
        return options_selected, criterion_feedback

    def reset_ora_test_data(self, course_id, submissions_config):
        """
        Reset all mentioned submitters in all mentioned ORAs by calling clear_student_state
        """
        learner_usernames, _ = self.get_usernames(submissions_config)
        display_names = self.get_display_names(submissions_config)

        for display_name in display_names:
            block = self.display_name_to_block[display_name]
            for learner_username in learner_usernames:
                log.info("Resetting learner state for user %s ORA '%s'", learner_username, display_name)
                block.clear_student_state(
                    self.username_to_anonymous_user_id[learner_username],
                    str(course_id),
                    str(block.location),
                    self.username_to_anonymous_user_id[SUPERUSER_USERNAME],
                )



""" Urls for fileupload app. """

from django.urls import re_path

from openassessment.fileupload import views_django_storage, views_filesystem

urlpatterns = [
    re_path(r'^django/(?P<key>.+)/$', views_django_storage.django_storage,
            name='openassessment-django-storage'),
    re_path(r'^(?P<key>.+)/$', views_filesystem.filesystem_storage,
            name='openassessment-filesystem-storage'),
]



# pylint: skip-file


from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('assessment', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='StaffWorkflow',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('scorer_id', models.CharField(max_length=40, db_index=True)),
                ('course_id', models.CharField(max_length=40, db_index=True)),
                ('item_id', models.CharField(max_length=128, db_index=True)),
                ('submission_uuid', models.CharField(unique=True, max_length=128, db_index=True)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, db_index=True)),
                ('grading_completed_at', models.DateTimeField(null=True, db_index=True)),
                ('grading_started_at', models.DateTimeField(null=True, db_index=True)),
                ('cancelled_at', models.DateTimeField(null=True, db_index=True)),
                ('assessment', models.CharField(max_length=128, null=True, db_index=True)),
            ],
            options={
                'ordering': ['created_at', 'id'],
            },
        ),
    ]



# Generated by Django 2.2.24 on 2021-11-01 16:26

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SubmissionGradingLock',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('submission_uuid', models.CharField(db_index=True, max_length=128, unique=True)),
                ('owner_id', models.CharField(max_length=40)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
            ],
            options={
                'ordering': ['created_at', 'id'],
            },
        ),
    ]
